# -*- coding: utf-8 -*-

from .platform import Platform
from .keywords.postgresql_keywords import PostgreSQLKeywords
from ..table import Table
from ..column import Column
from ..identifier import Identifier


class PostgresPlatform(Platform):

    INTERNAL_TYPE_MAPPING = {
        "smallint": "smallint",
        "int2": "smallint",
        "serial": "integer",
        "serial4": "integer",
        "int": "integer",
        "int4": "integer",
        "integer": "integer",
        "bigserial": "bigint",
        "serial8": "bigint",
        "bigint": "bigint",
        "int8": "bigint",
        "bool": "boolean",
        "boolean": "boolean",
        "text": "text",
        "tsvector": "text",
        "varchar": "string",
        "interval": "string",
        "_varchar": "string",
        "char": "string",
        "bpchar": "string",
        "inet": "string",
        "date": "date",
        "datetime": "datetime",
        "timestamp": "datetime",
        "timestamptz": "datetimez",
        "time": "time",
        "timetz": "time",
        "float": "float",
        "float4": "float",
        "float8": "float",
        "double": "float",
        "double precision": "float",
        "real": "float",
        "decimal": "decimal",
        "money": "decimal",
        "numeric": "decimal",
        "year": "date",
        "uuid": "guid",
        "bytea": "blob",
        "json": "json",
    }

    def get_list_table_columns_sql(self, table):
        sql = """SELECT
                    a.attnum,
                    quote_ident(a.attname) AS field,
                    t.typname AS type,
                    format_type(a.atttypid, a.atttypmod) AS complete_type,
                    (SELECT t1.typname FROM pg_catalog.pg_type t1 WHERE t1.oid = t.typbasetype) AS domain_type,
                    (SELECT format_type(t2.typbasetype, t2.typtypmod) FROM
                      pg_catalog.pg_type t2 WHERE t2.typtype = 'd' AND t2.oid = a.atttypid) AS domain_complete_type,
                    a.attnotnull AS isnotnull,
                    (SELECT 't'
                     FROM pg_index
                     WHERE c.oid = pg_index.indrelid
                        AND pg_index.indkey[0] = a.attnum
                        AND pg_index.indisprimary = 't'
                    ) AS pri,
                    (SELECT pg_get_expr(adbin, adrelid)
                     FROM pg_attrdef
                     WHERE c.oid = pg_attrdef.adrelid
                        AND pg_attrdef.adnum=a.attnum
                    ) AS default,
                    (SELECT pg_description.description
                        FROM pg_description WHERE pg_description.objoid = c.oid AND a.attnum = pg_description.objsubid
                    ) AS comment
                    FROM pg_attribute a, pg_class c, pg_type t, pg_namespace n
                    WHERE %s
                        AND a.attnum > 0
                        AND a.attrelid = c.oid
                        AND a.atttypid = t.oid
                        AND n.oid = c.relnamespace
                    ORDER BY a.attnum""" % self.get_table_where_clause(
            table
        )

        return sql

    def get_list_table_indexes_sql(self, table):
        sql = """
              SELECT quote_ident(relname) as relname, pg_index.indisunique, pg_index.indisprimary,
                     pg_index.indkey, pg_index.indrelid,
                     pg_get_expr(indpred, indrelid) AS where
              FROM pg_class, pg_index
              WHERE oid IN (
                  SELECT indexrelid
                  FROM pg_index si, pg_class sc, pg_namespace sn
                  WHERE %s
                  AND sc.oid=si.indrelid AND sc.relnamespace = sn.oid
              ) AND pg_index.indexrelid = oid"""

        sql = sql % self.get_table_where_clause(table, "sc", "sn")

        return sql

    def get_list_table_foreign_keys_sql(self, table):
        return (
            "SELECT quote_ident(r.conname) as conname, "
            "pg_catalog.pg_get_constraintdef(r.oid, true) AS condef "
            "FROM pg_catalog.pg_constraint r "
            "WHERE r.conrelid = "
            "("
            "SELECT c.oid "
            "FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n "
            "WHERE "
            + self.get_table_where_clause(table)
            + " AND n.oid = c.relnamespace"
            ")"
            " AND r.contype = 'f'"
        )

    def get_table_where_clause(self, table, class_alias="c", namespace_alias="n"):
        where_clause = (
            namespace_alias
            + ".nspname NOT IN ('pg_catalog', 'information_schema', 'pg_toast') AND "
        )
        if table.find(".") >= 0:
            split = table.split(".")
            schema, table = split[0], split[1]
            schema = "'%s'" % schema
        else:
            schema = (
                "ANY(string_to_array((select replace(replace(setting, '\"$user\"', user), ' ', '')"
                " from pg_catalog.pg_settings where name = 'search_path'),','))"
            )

        where_clause += "%s.relname = '%s' AND %s.nspname = %s" % (
            class_alias,
            table,
            namespace_alias,
            schema,
        )

        return where_clause

    def get_advanced_foreign_key_options_sql(self, foreign_key):
        query = ""

        if foreign_key.has_option("match"):
            query += " MATCH %s" % foreign_key.get_option("match")

        query += super(PostgresPlatform, self).get_advanced_foreign_key_options_sql(
            foreign_key
        )

        deferrable = (
            foreign_key.has_option("deferrable")
            and foreign_key.get_option("deferrable") is not False
        )
        if deferrable:
            query += " DEFERRABLE"
        else:
            query += " NOT DEFERRABLE"

        query += " INITIALLY"

        deferred = (
            foreign_key.has_option("deferred")
            and foreign_key.get_option("deferred") is not False
        )
        if deferred:
            query += " DEFERRED"
        else:
            query += " IMMEDIATE"

        return query

    def get_alter_table_sql(self, diff):
        """
        Get the ALTER TABLE SQL statement

        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: list
        """
        sql = []

        for column_diff in diff.changed_columns.values():
            if self.is_unchanged_binary_column(column_diff):
                continue

            old_column_name = column_diff.get_old_column_name().get_quoted_name(self)
            column = column_diff.column

            if any(
                [
                    column_diff.has_changed("type"),
                    column_diff.has_changed("precision"),
                    column_diff.has_changed("scale"),
                    column_diff.has_changed("fixed"),
                ]
            ):
                query = (
                    "ALTER "
                    + old_column_name
                    + " TYPE "
                    + self.get_sql_type_declaration(column.to_dict())
                )
                sql.append(
                    "ALTER TABLE "
                    + diff.get_name(self).get_quoted_name(self)
                    + " "
                    + query
                )

            if column_diff.has_changed("default") or column_diff.has_changed("type"):
                if column.get_default() is None:
                    default_clause = " DROP DEFAULT"
                else:
                    default_clause = " SET" + self.get_default_value_declaration_sql(
                        column.to_dict()
                    )

                query = "ALTER " + old_column_name + default_clause
                sql.append(
                    "ALTER TABLE "
                    + diff.get_name(self).get_quoted_name(self)
                    + " "
                    + query
                )

            if column_diff.has_changed("notnull"):
                op = "DROP"
                if column.get_notnull():
                    op = "SET"

                query = "ALTER " + old_column_name + " " + op + " NOT NULL"
                sql.append(
                    "ALTER TABLE "
                    + diff.get_name(self).get_quoted_name(self)
                    + " "
                    + query
                )

            if column_diff.has_changed("autoincrement"):
                if column.get_autoincrement():
                    seq_name = self.get_identity_sequence_name(
                        diff.name, old_column_name
                    )

                    sql.append("CREATE SEQUENCE " + seq_name)
                    sql.append(
                        "SELECT setval('" + seq_name + "', "
                        "(SELECT MAX(" + old_column_name + ") FROM " + diff.name + "))"
                    )
                    query = (
                        "ALTER "
                        + old_column_name
                        + " SET DEFAULT nextval('"
                        + seq_name
                        + "')"
                    )
                    sql.append(
                        "ALTER TABLE "
                        + diff.get_name(self).get_quoted_name(self)
                        + " "
                        + query
                    )
                else:
                    query = "ALTER " + old_column_name + " DROP DEFAULT"
                    sql.append(
                        "ALTER TABLE "
                        + diff.get_name(self).get_quoted_name(self)
                        + " "
                        + query
                    )

            if column_diff.has_changed("length"):
                query = (
                    "ALTER "
                    + old_column_name
                    + " TYPE "
                    + self.get_sql_type_declaration(column.to_dict())
                )
                sql.append(
                    "ALTER TABLE "
                    + diff.get_name(self).get_quoted_name(self)
                    + " "
                    + query
                )

        for old_column_name, column in diff.renamed_columns.items():
            sql.append(
                "ALTER TABLE " + diff.get_name(self).get_quoted_name(self) + " "
                "RENAME COLUMN "
                + Identifier(old_column_name).get_quoted_name(self)
                + " TO "
                + column.get_quoted_name(self)
            )

        return sql

    def is_unchanged_binary_column(self, column_diff):
        column_type = column_diff.column.get_type()

        if column_type not in ["blob", "binary"]:
            return False

        if isinstance(column_diff.from_column, Column):
            from_column = column_diff.from_column
        else:
            from_column = None

        if from_column:
            from_column_type = self.INTERNAL_TYPE_MAPPING[from_column.get_type()]

            if from_column_type in ["blob", "binary"]:
                return False

            return (
                len(
                    [
                        x
                        for x in column_diff.changed_properties
                        if x not in ["type", "length", "fixed"]
                    ]
                )
                == 0
            )

        if column_diff.has_changed("type"):
            return False

        return (
            len(
                [
                    x
                    for x in column_diff.changed_properties
                    if x not in ["length", "fixed"]
                ]
            )
            == 0
        )

    def convert_booleans(self, item):
        if isinstance(item, list):
            for i, value in enumerate(item):
                if isinstance(value, bool):
                    item[i] = str(value).lower()
        elif isinstance(item, bool):
            item = str(item).lower()

        return item

    def get_boolean_type_declaration_sql(self, column):
        return "BOOLEAN"

    def get_integer_type_declaration_sql(self, column):
        if column.get("autoincrement"):
            return "SERIAL"

        return "INT"

    def get_bigint_type_declaration_sql(self, column):
        if column.get("autoincrement"):
            return "BIGSERIAL"

        return "BIGINT"

    def get_smallint_type_declaration_sql(self, column):
        return "SMALLINT"

    def get_guid_type_declaration_sql(self, column):
        return "UUID"

    def get_datetime_type_declaration_sql(self, column):
        return "TIMESTAMP(0) WITHOUT TIME ZONE"

    def get_datetimetz_type_declaration_sql(self, column):
        return "TIMESTAMP(0) WITH TIME ZONE"

    def get_date_type_declaration_sql(self, column):
        return "DATE"

    def get_time_type_declaration_sql(self, column):
        return "TIME(0) WITHOUT TIME ZONE"

    def get_string_type_declaration_sql(self, column):
        length = column.get("length", "255")
        fixed = column.get("fixed")

        if fixed:
            return "CHAR(%s)" % length
        else:
            return "VARCHAR(%s)" % length

    def get_binary_type_declaration_sql(self, column):
        return "BYTEA"

    def get_blob_type_declaration_sql(self, column):
        return "BYTEA"

    def get_clob_type_declaration_sql(self, column):
        return "TEXT"

    def get_text_type_declaration_sql(self, column):
        return "TEXT"

    def get_json_type_declaration_sql(self, column):
        return "JSON"

    def get_decimal_type_declaration_sql(self, column):
        if "precision" not in column or not column["precision"]:
            column["precision"] = 10

        if "scale" not in column or not column["scale"]:
            column["precision"] = 0

        return "DECIMAL(%s, %s)" % (column["precision"], column["scale"])

    def get_float_type_declaration_sql(self, column):
        return "DOUBLE PRECISION"

    def supports_foreign_key_constraints(self):
        return True

    def has_native_json_type(self):
        return True

    def _get_reserved_keywords_class(self):
        return PostgreSQLKeywords



# -*- coding: utf-8 -*-

from .database_migration_repository import DatabaseMigrationRepository
from .migration_creator import MigrationCreator
from .migration import Migration
from .migrator import Migrator



# -*- coding: utf-8 -*-


class Scope(object):
    def apply(self, builder, model):
        """
        Apply the scope to a given query builder.

        :param builder: The query builder
        :type builder: orator.orm.Builder

        :param model: The model
        :type model: orator.orm.Model
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

import simplejson as json
import pendulum
import inflection
import inspect
import uuid
import datetime
from warnings import warn
from six import add_metaclass
from collections import OrderedDict
from ..utils import basestring, deprecated
from ..exceptions.orm import MassAssignmentError, RelatedClassNotFound
from ..query import QueryBuilder
from .builder import Builder
from .collection import Collection
from .relations import (
    Relation,
    HasOne,
    HasMany,
    BelongsTo,
    BelongsToMany,
    HasManyThrough,
    MorphOne,
    MorphMany,
    MorphTo,
    MorphToMany,
)
from .relations.wrapper import Wrapper, BelongsToManyWrapper
from .utils import mutator, accessor
from .scopes import Scope
from ..events import Event


class ModelRegister(dict):
    def __init__(self, *args, **kwargs):
        self.inverse = {}

        super(ModelRegister, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super(ModelRegister, self).__setitem__(key, value)

        self.inverse[value] = key

    def __delitem__(self, key):
        del self.inverse[self[key]]

        super(ModelRegister, self).__delitem__(key)


class MetaModel(type):

    __register__ = {}

    def __init__(cls, *args, **kwargs):
        name = cls.__table__ or inflection.tableize(cls.__name__)
        cls._register[name] = cls

        super(MetaModel, cls).__init__(*args, **kwargs)

    def __getattr__(cls, item):
        try:
            return type.__getattribute__(cls, item)
        except AttributeError:
            query = cls.query()

            return getattr(query, item)


@add_metaclass(MetaModel)
class Model(object):

    __connection__ = None

    __table__ = None

    __primary_key__ = "id"

    __incrementing__ = True

    __fillable__ = []
    __guarded__ = ["*"]
    __unguarded__ = False

    __hidden__ = []
    __visible__ = []
    __appends__ = []

    __timestamps__ = True
    __dates__ = []

    __casts__ = {}

    __touches__ = []

    __morph_name__ = None

    _per_page = 15

    _with = []

    _booted = {}
    _global_scopes = {}
    _registered = []

    _accessor_cache = {}
    _mutator_cache = {}

    __resolver = None
    __columns__ = []

    __dispatcher__ = Event()
    __observables__ = []

    _register = ModelRegister()

    __attributes__ = {}

    many_methods = ["belongs_to_many", "morph_to_many", "morphed_by_many"]

    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"

    def __init__(self, _attributes=None, **attributes):
        """
        :param attributes: The instance attributes
        """
        self._boot_if_not_booted()

        self._exists = False
        self._original = {}

        # Setting default attributes' values
        self._attributes = dict((k, v) for k, v in self.__attributes__.items())
        self._relations = {}

        self.sync_original()

        if _attributes is not None:
            attributes.update(_attributes)

        self.fill(**attributes)

    def _boot_if_not_booted(self):
        """
        Check if the model needs to be booted and if so, do it.
        """
        klass = self.__class__

        if not klass._booted.get(klass):
            klass._booted[klass] = True

            self._fire_model_event("booting")

            klass._boot()

            self._fire_model_event("booted")

    @classmethod
    def _boot(cls):
        """
        The booting method of the model.
        """
        cls._accessor_cache[cls] = {}
        cls._mutator_cache[cls] = {}

        for name, method in cls.__dict__.items():
            if isinstance(method, accessor):
                cls._accessor_cache[cls][method.attribute] = method
            elif isinstance(method, mutator):
                cls._mutator_cache[cls][method.attribute] = method

        cls._boot_mixins()

    @classmethod
    def _boot_columns(cls):
        connection = cls.resolve_connection()
        columns = connection.get_schema_manager().list_table_columns(
            cls.__table__ or inflection.tableize(cls.__name__)
        )
        cls.__columns__ = list(columns.keys())

    @classmethod
    def _boot_mixins(cls):
        """
        Boot the mixins
        """
        for mixin in cls.__bases__:
            # if mixin == Model:
            #    continue

            method = "boot_%s" % inflection.underscore(mixin.__name__)
            if hasattr(mixin, method):
                getattr(mixin, method)(cls)

    @classmethod
    def add_global_scope(cls, scope, implementation=None):
        """
        Register a new global scope on the model.

        :param scope: The scope to register
        :type scope: orator.orm.scopes.scope.Scope or callable or str

        :param implementation: The scope implementation
        :type implementation: callbale or None
        """
        if cls not in cls._global_scopes:
            cls._global_scopes[cls] = OrderedDict()

        if isinstance(scope, basestring) and implementation is not None:
            cls._global_scopes[cls][scope] = implementation
        elif callable(scope):
            cls._global_scopes[cls][uuid.uuid4().hex] = scope
        elif isinstance(scope, Scope):
            cls._global_scopes[cls][scope.__class__] = scope
        else:
            raise Exception("Global scope must be an instance of Scope or a callable")

    @classmethod
    def has_global_scope(cls, scope):
        """
        Determine if a model has a global scope.

        :param scope: The scope to register
        :type scope: orator.orm.scopes.scope.Scope or str
        """
        return cls.get_global_scope(scope) is not None

    @classmethod
    def get_global_scope(cls, scope):
        """
        Get a global scope registered with the model.

        :param scope: The scope to register
        :type scope: orator.orm.scopes.scope.Scope or str
        """
        for key, value in cls._global_scopes[cls].items():
            if isinstance(scope, key):
                return value

    def get_global_scopes(self):
        """
        Get the global scopes for this class instance.

        :rtype: dict
        """
        return self.__class__._global_scopes.get(self.__class__, {})

    @classmethod
    def observe(cls, observer):
        """
        Register an observer with the Model.

        :param observer: The observer
        """
        for event in cls.get_observable_events():
            if hasattr(observer, event):
                cls._register_model_event(event, getattr(observer, event))

    def fill(self, _attributes=None, **attributes):
        """
        Fill the model with attributes.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The model instance
        :rtype: Model

        :raises: MassAssignmentError
        """
        if _attributes is not None:
            attributes.update(_attributes)

        totally_guarded = self.totally_guarded()

        for key, value in self._fillable_from_dict(attributes).items():
            key = self._remove_table_from_key(key)

            if self.is_fillable(key):
                self.set_attribute(key, value)
            elif totally_guarded:
                raise MassAssignmentError(key)

        return self

    def force_fill(self, _attributes=None, **attributes):
        """
        Fill the model with attributes. Force mass assignment.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The model instance
        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        self.unguard()

        self.fill(**attributes)

        self.reguard()

        return self

    def _fillable_from_dict(self, attributes):
        """
        Get the fillable attributes from a given dictionary.

        :type attributes: dict

        :return: The fillable attributes
        :rtype: dict
        """
        if self.__fillable__ and not self.__unguarded__:
            return {x: attributes[x] for x in attributes if x in self.__fillable__}

        return attributes

    def new_instance(self, attributes=None, exists=False):
        """
        Create a new instance for the given model.

        :param attributes: The instance attributes
        :type attributes: dict

        :param exists:
        :type exists: bool

        :return: A new instance for the current model
        :rtype: Model
        """
        if attributes is None:
            attributes = {}

        model = self.__class__(**attributes)

        model.set_connection(self.get_connection_name())
        model.set_exists(exists)

        return model

    def new_from_builder(self, attributes=None, connection=None):
        """
        Create a new model instance that is existing.

        :param attributes: The model attributes
        :type attributes: dict

        :param connection: The connection name
        :type connection: str

        :return: A new instance for the current model
        :rtype: Model
        """
        model = self.new_instance({}, True)

        if attributes is None:
            attributes = {}

        model.set_raw_attributes(attributes, True)

        model.set_connection(connection or self.__connection__)

        return model

    @classmethod
    def hydrate(cls, items, connection=None):
        """
        Create a collection of models from plain lists.

        :param items:
        :param connection:
        :return:
        """
        instance = cls().set_connection(connection)

        collection = instance.new_collection(items)

        return collection.map(lambda item: instance.new_from_builder(item))

    @classmethod
    def hydrate_raw(cls, query, bindings=None, connection=None):
        """
        Create a collection of models from a raw query.

        :param query: The SQL query
        :type query: str

        :param bindings: The query bindings
        :type bindings: list

        :param connection: The connection name

        :rtype: Collection
        """
        instance = cls().set_connection(connection)

        items = instance.get_connection().select(query, bindings)

        return cls.hydrate(items, connection)

    @classmethod
    def create(cls, _attributes=None, **attributes):
        """
        Save a new model an return the instance.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        model = cls(**attributes)

        model.save()

        return model

    @classmethod
    def force_create(cls, **attributes):
        """
        Save a new model an return the instance. Allow mass assignment.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        cls.unguard()

        model = cls.create(**attributes)

        cls.reguard()

        return model

    @classmethod
    def first_or_create(cls, **attributes):
        """
        Get the first record matching the attributes or create it.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls().new_query_without_scopes().where(attributes).first()

        if instance is not None:
            return instance

        return cls.create(**attributes)

    @classmethod
    def first_or_new(cls, **attributes):
        """
        Get the first record matching the attributes or instantiate it.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls().new_query_without_scopes().where(attributes).first()

        if instance is not None:
            return instance

        return cls(**attributes)

    @classmethod
    def update_or_create(cls, attributes, values=None):
        """
        Create or update a record matching the attributes, and fill it with values.

        :param attributes: The instance attributes
        :type attributes: dict

        :param values: The values
        :type values: dict

        :return: The new instance
        :rtype: Model
        """
        instance = cls.first_or_new(**attributes)

        if values is None:
            values = {}

        instance.fill(**values).save()

        return instance

    @classmethod
    def query(cls):
        """
        Begin querying the model.

        :return: A Builder instance
        :rtype: orator.orm.Builder
        """
        return cls().new_query()

    @classmethod
    def on(cls, connection=None):
        """
        Begin querying the model on a given connection.

        :param connection: The connection name
        :type connection: str

        :return: A Builder instance
        :rtype: orator.orm.Builder
        """
        instance = cls()

        instance.set_connection(connection)

        return instance.new_query()

    @classmethod
    def on_write_connection(cls):
        """
        Begin querying the model on the write connection.

        :return: A Builder instance
        :rtype: QueryBuilder
        """
        instance = cls()

        return instance.new_query().use_write_connection()

    @classmethod
    def all(cls, columns=None):
        """
        Get all og the models from the database.

        :param columns: The columns to retrieve
        :type columns: list

        :return: A Collection instance
        :rtype: Collection
        """
        instance = cls()

        return instance.new_query().get(columns)

    @classmethod
    def find(cls, id, columns=None):
        """
        Find a model by its primary key.

        :param id: The id of the model
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: Either a Model instance or a Collection
        :rtype: Model
        """
        instance = cls()

        if isinstance(id, list) and not id:
            return instance.new_collection()

        if columns is None:
            columns = ["*"]

        return instance.new_query().find(id, columns)

    @classmethod
    def find_or_new(cls, id, columns=None):
        """
        Find a model by its primary key or return new instance.

        :param id: The id of the model
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: A Model instance
        :rtype: Model
        """
        instance = cls.find(id, columns)

        if instance is not None:
            return instance

        return cls()

    def fresh(self, with_=None):
        """
        Reload a fresh instance from the database.

        :param with_: The list of relations to eager load
        :type with_: list

        :return: The current model instance
        :rtype: Model
        """
        if with_ is None:
            with_ = ()

        key = self.get_key_name()

        if self.exists:
            return self.with_(*with_).where(key, self.get_key()).first()

    def load(self, *relations):
        """
        Eager load relations on the model

        :param relations: The relations to eager load
        :type relations: str or list

        :return: The current model instance
        :rtype: Model
        """
        query = self.new_query().with_(*relations)

        query.eager_load_relations([self])

        return self

    @classmethod
    def with_(cls, *relations):
        """
        Begin querying a model with eager loading

        :param relations: The relations to eager load
        :type relations: str or list

        :return: A Builder instance
        :rtype: Builder
        """
        instance = cls()

        return instance.new_query().with_(*relations)

    def has_one(
        self, related, foreign_key=None, local_key=None, relation=None, _wrapped=True
    ):
        """
        Define a one to one relationship.

        :param related: The related model:
        :type related: Model or str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param local_key: The local key
        :type local_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: HasOne
        """
        if relation is None:
            name = inspect.stack()[1][3]
        else:
            name = relation

        if name in self._relations:
            return self._relations[name]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = self._get_related(related, True)

        if not local_key:
            local_key = self.get_key_name()

        rel = HasOne(
            instance.new_query(),
            self,
            "%s.%s" % (instance.get_table(), foreign_key),
            local_key,
        )

        if _wrapped:
            warn(
                "Using has_one method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[name] = rel

        return rel

    def morph_one(
        self,
        related,
        name,
        type_column=None,
        id_column=None,
        local_key=None,
        relation=None,
        _wrapped=True,
    ):
        """
        Define a polymorphic one to one relationship.

        :param related: The related model:
        :type related: Model or str

        :param type_column: The name of the type column
        :type type_column: str

        :param id_column: The name of the id column
        :type id_column: str

        :param local_key: The local key
        :type local_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: HasOne
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self._relations:
            return self._relations[name]

        instance = self._get_related(related, True)

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        table = instance.get_table()

        if not local_key:
            local_key = self.get_key_name()

        rel = MorphOne(
            instance.new_query(),
            self,
            "%s.%s" % (table, type_column),
            "%s.%s" % (table, id_column),
            local_key,
        )

        if _wrapped:
            warn(
                "Using morph_one method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[relation] = rel

        return rel

    def belongs_to(
        self, related, foreign_key=None, other_key=None, relation=None, _wrapped=True
    ):
        """
        Define an inverse one to one or many relationship.

        :param related: The related model:
        :type related: Model or str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :type relation: str

        :rtype: BelongsTo
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self._relations:
            return self._relations[relation]

        if foreign_key is None:
            foreign_key = "%s_id" % inflection.underscore(relation)

        instance = self._get_related(related, True)

        query = instance.new_query()

        if not other_key:
            other_key = instance.get_key_name()

        rel = BelongsTo(query, self, foreign_key, other_key, relation)

        if _wrapped:
            warn(
                "Using belongs_to method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[relation] = rel

        return rel

    def morph_to(self, name=None, type_column=None, id_column=None, _wrapped=True):
        """
        Define a polymorphic, inverse one-to-one or many relationship.

        :param name: The name of the relation
        :type name: str

        :param type_column: The type column
        :type type_column: str

        :param id_column: The id column
        :type id_column: str

        :rtype: MorphTo
        """
        if not name:
            name = inspect.stack()[1][3]

        if name in self._relations:
            return self._relations[name]

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        # If the type value is null it is probably safe to assume we're eager loading
        # the relationship. When that is the case we will pass in a dummy query as
        # there are multiple types in the morph and we can't use single queries.
        if not hasattr(self, type_column):
            return MorphTo(self.new_query(), self, id_column, None, type_column, name)

        # If we are not eager loading the relationship we will essentially treat this
        # as a belongs-to style relationship since morph-to extends that class and
        # we will pass in the appropriate values so that it behaves as expected.
        klass = self.get_actual_class_for_morph(getattr(self, type_column))

        instance = klass()
        instance.set_connection(self.get_connection_name())

        rel = MorphTo(
            instance.new_query(),
            self,
            id_column,
            instance.get_key_name(),
            type_column,
            name,
        )

        if _wrapped:
            warn(
                "Using morph_to method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[name] = rel

        return rel

    def get_actual_class_for_morph(self, slug):
        """
        Retrieve the class from a slug.

        :param slug: The slug
        :type slug: str
        """
        for cls in self.__class__._register.values():
            morph_name = cls.get_morph_name()
            if morph_name == slug:
                return cls

    def has_many(
        self, related, foreign_key=None, local_key=None, relation=None, _wrapped=True
    ):
        """
        Define a one to many relationship.

        :param related: The related model
        :type related: Model or str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param local_key: The local key
        :type local_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: HasOne
        """
        if relation is None:
            name = inspect.stack()[1][3]
        else:
            name = relation

        if name in self._relations:
            return self._relations[name]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = self._get_related(related, True)

        if not local_key:
            local_key = self.get_key_name()

        rel = HasMany(
            instance.new_query(),
            self,
            "%s.%s" % (instance.get_table(), foreign_key),
            local_key,
        )

        if _wrapped:
            warn(
                "Using has_many method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[name] = rel

        return rel

    def has_many_through(
        self,
        related,
        through,
        first_key=None,
        second_key=None,
        relation=None,
        _wrapped=True,
    ):
        """
        Define a has-many-through relationship.

        :param related: The related model
        :type related: Model or str

        :param through: The through model
        :type through: Model or str

        :param first_key: The first key
        :type first_key: str

        :param second_key: The second_key
        :type second_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: HasManyThrough
        """
        if relation is None:
            name = inspect.stack()[1][3]
        else:
            name = relation

        if name in self._relations:
            return self._relations[name]

        through = self._get_related(through, True)

        if not first_key:
            first_key = self.get_foreign_key()

        if not second_key:
            second_key = through.get_foreign_key()

        rel = HasManyThrough(
            self._get_related(related)().new_query(),
            self,
            through,
            first_key,
            second_key,
        )

        if _wrapped:
            warn(
                "Using has_many_through method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[name] = rel

        return rel

    def morph_many(
        self,
        related,
        name,
        type_column=None,
        id_column=None,
        local_key=None,
        relation=None,
        _wrapped=True,
    ):
        """
        Define a polymorphic one to many relationship.

        :param related: The related model:
        :type related: Model or str

        :param type_column: The name of the type column
        :type type_column: str

        :param id_column: The name of the id column
        :type id_column: str

        :param local_key: The local key
        :type local_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: MorphMany
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self._relations:
            return self._relations[relation]

        instance = self._get_related(related, True)

        type_column, id_column = self.get_morphs(name, type_column, id_column)

        table = instance.get_table()

        if not local_key:
            local_key = self.get_key_name()

        rel = MorphMany(
            instance.new_query(),
            self,
            "%s.%s" % (table, type_column),
            "%s.%s" % (table, id_column),
            local_key,
        )

        if _wrapped:
            warn(
                "Using morph_many method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[name] = rel

        return rel

    def belongs_to_many(
        self,
        related,
        table=None,
        foreign_key=None,
        other_key=None,
        relation=None,
        _wrapped=True,
    ):
        """
        Define a many-to-many relationship.

        :param related: The related model
        :type related: Model or str

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :type relation: str

        :rtype: BelongsToMany
        """
        if relation is None:
            relation = inspect.stack()[1][3]

        if relation in self._relations:
            return self._relations[relation]

        if not foreign_key:
            foreign_key = self.get_foreign_key()

        instance = self._get_related(related, True)

        if not other_key:
            other_key = instance.get_foreign_key()

        if table is None:
            table = self.joining_table(instance)

        query = instance.new_query()

        rel = BelongsToMany(query, self, table, foreign_key, other_key, relation)

        if _wrapped:
            warn(
                "Using belongs_to_many method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = BelongsToManyWrapper(rel)

        self._relations[relation] = rel

        return rel

    def morph_to_many(
        self,
        related,
        name,
        table=None,
        foreign_key=None,
        other_key=None,
        inverse=False,
        relation=None,
        _wrapped=True,
    ):
        """
        Define a polymorphic many-to-many relationship.

        :param related: The related model:
        :type related: Model or str

        :param name: The relation name
        :type name: str

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: MorphToMany
        """
        if relation is None:
            caller = inspect.stack()[1][3]
        else:
            caller = relation

        if caller in self._relations:
            return self._relations[caller]

        if not foreign_key:
            foreign_key = name + "_id"

        instance = self._get_related(related, True)

        if not other_key:
            other_key = instance.get_foreign_key()

        query = instance.new_query()

        if not table:
            table = inflection.pluralize(name)

        rel = MorphToMany(
            query, self, name, table, foreign_key, other_key, caller, inverse
        )

        if _wrapped:
            warn(
                "Using morph_to_many method directly is deprecated. "
                "Use the appropriate decorator instead.",
                category=DeprecationWarning,
            )

            rel = Wrapper(rel)

        self._relations[caller] = rel

        return rel

    def morphed_by_many(
        self,
        related,
        name,
        table=None,
        foreign_key=None,
        other_key=None,
        relation=None,
        _wrapped=False,
    ):
        """
        Define a polymorphic many-to-many relationship.

        :param related: The related model:
        :type related: Model or str

        :param name: The relation name
        :type name: str

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :param relation: The name of the relation (defaults to method name)
        :type relation: str

        :rtype: MorphToMany
        """
        if not foreign_key:
            foreign_key = self.get_foreign_key()

        if not other_key:
            other_key = name + "_id"

        return self.morph_to_many(
            related, name, table, foreign_key, other_key, True, relation, _wrapped
        )

    def _get_related(self, related, as_instance=False):
        """
        Get the related class.

        :param related: The related model or table
        :type related: Model or str

        :rtype: Model class
        """
        if not isinstance(related, basestring) and issubclass(related, Model):
            if as_instance:
                instance = related()
                instance.set_connection(self.get_connection_name())

                return instance

            return related

        related_class = self.__class__._register.get(related)

        if related_class:
            if as_instance:
                instance = related_class()
                instance.set_connection(self.get_connection_name())

                return instance

            return related_class

        raise RelatedClassNotFound(related)

    def joining_table(self, related):
        """
        Get the joining table name for a many-to-many relation

        :param related: The related model
        :type related: Model

        :rtype: str
        """
        base = self.get_table()

        related = related.get_table()

        models = sorted([related, base])

        return "_".join(models)

    @classmethod
    def destroy(cls, *ids):
        """
        Destroy the models for the given IDs

        :param ids: The ids of the models to destroy
        :type ids: tuple

        :return: The number of models destroyed
        :rtype: int
        """
        count = 0

        if len(ids) == 1 and isinstance(ids[0], list):
            ids = ids[0]

        ids = list(ids)

        instance = cls()

        key = instance.get_key_name()

        for model in instance.new_query().where_in(key, ids).get():
            if model.delete():
                count += 1

        return count

    def delete(self):
        """
        Delete the model from the database.

        :rtype: bool or None

        :raises: Exception
        """
        if self.__primary_key__ is None:
            raise Exception("No primary key defined on the model.")

        if self._exists:
            if self._fire_model_event("deleting") is False:
                return False

            self.touch_owners()

            self._perform_delete_on_model()

            self._exists = False

            self._fire_model_event("deleted")

            return True

    def force_delete(self):
        """
        Force a hard delete on a soft deleted model.
        """
        return self.delete()

    def _perform_delete_on_model(self):
        """
        Perform the actual delete query on this model instance.
        """
        if hasattr(self, "_do_perform_delete_on_model"):
            return self._do_perform_delete_on_model()

        return self.new_query().where(self.get_key_name(), self.get_key()).delete()

    @classmethod
    def saving(cls, callback):
        """
        Register a saving model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("saving", callback)

    @classmethod
    def saved(cls, callback):
        """
        Register a saved model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("saved", callback)

    @classmethod
    def updating(cls, callback):
        """
        Register a updating model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("updating", callback)

    @classmethod
    def updated(cls, callback):
        """
        Register a updated model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("updated", callback)

    @classmethod
    def creating(cls, callback):
        """
        Register a creating model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("creating", callback)

    @classmethod
    def created(cls, callback):
        """
        Register a created model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("created", callback)

    @classmethod
    def deleting(cls, callback):
        """
        Register a deleting model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("deleting", callback)

    @classmethod
    def deleted(cls, callback):
        """
        Register a deleted model event with the dispatcher.

        :type callback: callable
        """
        cls._register_model_event("deleted", callback)

    @classmethod
    def flush_event_listeners(cls):
        """
        Remove all of the event listeners for the model.
        """
        if not cls.__dispatcher__:
            return

        for event in cls.get_observable_events():
            cls.__dispatcher__.forget("%s: %s" % (event, cls.__name__))

    @classmethod
    def _register_model_event(cls, event, callback):
        """
        Register a model event with the dispatcher.

        :param event: The event
        :type event: str

        :param callback: The callback
        :type callback: callable
        """
        if cls.__dispatcher__:
            cls.__dispatcher__.listen("%s: %s" % (event, cls.__name__), callback)

    @classmethod
    def get_observable_events(cls):
        """
        Get the observable event names.

        :rtype: list
        """
        default_events = [
            "creating",
            "created",
            "updating",
            "updated",
            "deleting",
            "deleted",
            "saving",
            "saved",
            "restoring",
            "restored",
        ]

        return default_events + cls.__observables__

    def _increment(self, column, amount=1):
        """
        Increment a column's value

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :return: The new column value
        :rtype: int
        """
        return self._increment_or_decrement(column, amount, "increment")

    def _decrement(self, column, amount=1):
        """
        Decrement a column's value

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :return: The new column value
        :rtype: int
        """
        return self._increment_or_decrement(column, amount, "decrement")

    def _increment_or_decrement(self, column, amount, method):
        """
        Runthe increment or decrement method on the model

        :param column: The column to increment or decrement
        :type column: str

        :param amount: The amount by which to increment or decrement
        :type amount: int

        :param method: The method
        :type method: str

        :return: The new column value
        :rtype: int
        """
        query = self.new_query()

        if not self._exists:
            return getattr(query, method)(column, amount)

        self._increment_or_decrement_attribute_value(column, amount, method)

        query = query.where(self.get_key_name(), self.get_key())

        return getattr(query, method)(column, amount)

    def _increment_or_decrement_attribute_value(self, column, amount, method):
        """
        Increment the underlying attribute value and sync with original.

        :param column: The column to increment or decrement
        :type column: str

        :param amount: The amount by which to increment or decrement
        :type amount: int

        :param method: The method
        :type method: str

        :return: None
        """
        setattr(
            self,
            column,
            getattr(self, column) + (amount if method == "increment" else amount * -1),
        )

        self.sync_original_attribute(column)

    def update(self, _attributes=None, **attributes):
        """
        Update the model in the database.

        :param attributes: The model attributes
        :type attributes: dict

        :return: The number of rows affected
        :rtype: int
        """
        if _attributes is not None:
            attributes.update(_attributes)

        if not self._exists:
            return self.new_query().update(**attributes)

        return self.fill(**attributes).save()

    def push(self):
        """
        Save the model and all of its relationship.
        """
        if not self.save():
            return False

        for models in self._relations.values():
            if isinstance(models, Collection):
                models = models.all()
            else:
                models = [models]

            for model in models:
                if not model:
                    continue

                if not model.push():
                    return False

        return True

    def save(self, options=None):
        """
        Save the model to the database.
        """
        if options is None:
            options = {}

        query = self.new_query()

        if self._fire_model_event("saving") is False:
            return False

        if self._exists:
            saved = self._perform_update(query, options)
        else:
            saved = self._perform_insert(query, options)

        if saved:
            self._finish_save(options)

        return saved

    def _finish_save(self, options):
        """
        Finish processing on a successful save operation.
        """
        self._fire_model_event("saved")

        self.sync_original()

        if options.get("touch", True):
            self.touch_owners()

    def _perform_update(self, query, options=None):
        """
        Perform a model update operation.

        :param query: A Builder instance
        :type query: Builder

        :param options: Extra options
        :type options: dict
        """
        if options is None:
            options = {}

        dirty = self.get_dirty()

        if len(dirty):
            if self._fire_model_event("updating") is False:
                return False

            if self.__timestamps__ and options.get("timestamps", True):
                self._update_timestamps()

            dirty = self.get_dirty()

            if len(dirty):
                self._set_keys_for_save_query(query).update(dirty)

                self._fire_model_event("updated")

        return True

    def _perform_insert(self, query, options=None):
        """
        Perform a model update operation.

        :param query: A Builder instance
        :type query: Builder

        :param options: Extra options
        :type options: dict
        """
        if options is None:
            options = {}

        if self._fire_model_event("creating") is False:
            return False

        if self.__timestamps__ and options.get("timestamps", True):
            self._update_timestamps()

        attributes = self._attributes

        if self.__incrementing__:
            self._insert_and_set_id(query, attributes)
        else:
            query.insert(attributes)

        self._exists = True

        self._fire_model_event("created")

        return True

    def _insert_and_set_id(self, query, attributes):
        """
        Insert the given attributes and set the ID on the model.

        :param query: A Builder instance
        :type query: Builder

        :param attributes: The attributes to insert
        :type attributes: dict
        """
        key_name = self.get_key_name()

        id = query.insert_get_id(attributes, key_name)

        self.set_attribute(key_name, id)

    def touch_owners(self):
        """
        Touch the owning relations of the model.
        """
        for relation in self.__touches__:
            if hasattr(self, relation):
                _relation = getattr(self, relation)

                if _relation:
                    _relation.touch()
                    _relation.touch_owners()

    def touches(self, relation):
        """
        Determine if a model touches a given relation.

        :param relation: The relation to check.
        :type relation: str

        :rtype: bool
        """
        return relation in self.__touches__

    def _fire_model_event(self, event):
        """
        Fire the given event for the model.

        :type event: str
        """
        if not self.__dispatcher__:
            return True

        # We will append the names of the class to the event to distinguish it from
        # other model events that are fired, allowing us to listen on each model
        # event set individually instead of catching event for all the models.
        event = "%s: %s" % (event, self.__class__.__name__)

        return self.__dispatcher__.fire(event, self)

    def _set_keys_for_save_query(self, query):
        """
        Set the keys for a save update query.

        :param query: A Builder instance
        :type query: Builder

        :return: The Builder instance
        :rtype: Builder
        """
        query.where(self.get_key_name(), self._get_key_for_save_query())

        return query

    def _get_key_for_save_query(self):
        """
        Get the primary key value for a save query.
        """
        if self.get_key_name() in self._original:
            return self._original[self.get_key_name()]

        return self._attributes[self.get_key_name()]

    def touch(self):
        """
        Update the model's timestamps.

        :rtype: bool
        """
        if not self.__timestamps__:
            return False

        self._update_timestamps()

        return self.save()

    def _update_timestamps(self):
        """
        Update the model's timestamps.
        """
        time = self.fresh_timestamp()

        if not self.is_dirty(self.UPDATED_AT) and self._should_set_timestamp(
            self.UPDATED_AT
        ):
            self.set_updated_at(time)

        if (
            not self._exists
            and not self.is_dirty(self.CREATED_AT)
            and self._should_set_timestamp(self.CREATED_AT)
        ):
            self.set_created_at(time)

    def _should_set_timestamp(self, timestamp):
        """
        Determine if a timestamp should be set.

        :param timestamp: The timestamp to check
        :type timestamp: str

        :rtype: bool
        """
        if isinstance(self.__timestamps__, bool):
            return self.__timestamps__

        return timestamp in self.__timestamps__

    def set_created_at(self, value):
        """
        Set the value of the "created at" attribute.

        :param value: The value
        :type value: datetime
        """
        self.set_attribute(self.CREATED_AT, value)

    def set_updated_at(self, value):
        """
        Set the value of the "updated at" attribute.

        :param value: The value
        :type value: datetime
        """
        self.set_attribute(self.UPDATED_AT, value)

    def get_created_at_column(self):
        """
        Get the name of the "created at" column.

        :rtype: str
        """
        return self.CREATED_AT

    def get_updated_at_column(self):
        """
        Get the name of the "updated at" column.

        :rtype: str
        """
        return self.UPDATED_AT

    def fresh_timestamp(self):
        """
        Get a fresh timestamp for the model.

        :return: pendulum.Pendulum
        """
        return pendulum.utcnow()

    def fresh_timestamp_string(self):
        """
        Get a fresh timestamp string for the model.

        :return: str
        """
        return self.from_datetime(self.fresh_timestamp())

    def new_query(self):
        """
        Get a new query builder for the model's table

        :return: A Builder instance
        :rtype: Builder
        """
        builder = self.new_query_without_scopes()

        for identifier, scope in self.get_global_scopes().items():
            builder.with_global_scope(identifier, scope)

        return builder

    def new_query_without_scope(self, scope):
        """
        Get a new query builder for the model's table without a given scope

        :return: A Builder instance
        :rtype: Builder
        """
        builder = self.new_query()

        return builder.without_global_scope(scope)

    def new_query_without_scopes(self):
        """
        Get a new query builder without any scopes.

        :return: A Builder instance
        :rtype: Builder
        """
        builder = self.new_orm_builder(self._new_base_query_builder())

        return builder.set_model(self).with_(*self._with)

    @classmethod
    def query(cls):
        return cls().new_query()

    def new_orm_builder(self, query):
        """
        Create a new orm query builder for the model

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: A Builder instance
        :rtype: Builder
        """
        return Builder(query)

    def _new_base_query_builder(self):
        """
        Get a new query builder instance for the connection.

        :return: A QueryBuilder instance
        :rtype: QueryBuilder
        """
        conn = self.get_connection()

        return conn.query()

    def new_collection(self, models=None):
        """
        Create a new Collection instance.

        :param models: A list of models
        :type models: list

        :return: A new Collection instance
        :rtype: Collection
        """
        if models is None:
            models = []

        return Collection(models)

    def new_pivot(self, parent, attributes, table, exists):
        """
        Create a new pivot model instance.

        :param parent: The parent model
        :type parent: Model

        :param attributes: The pivot attributes
        :type attributes: dict

        :param table: the pivot table
        :type table: str

        :param exists: Whether the pivot exists or not
        :type exists: bool

        :rtype: Pivot
        """
        from .relations.pivot import Pivot

        return Pivot(parent, attributes, table, exists)

    def get_table(self):
        """
        Get the table associated with the model.

        :return: The name of the table
        :rtype: str
        """
        return self.__class__._register.inverse[self.__class__]

    def set_table(self, table):
        """
        Set the table associated with the model.

        :param table: The table name
        :type table: str
        """
        old_table = self.__class__._register.inverse.get(self.__class__, None)
        self.__table__ = table

        if old_table:
            del self.__class__._register[old_table]

        self.__class__._register[self.__table__] = self.__class__

    def get_key(self):
        """
        Get the value of the model's primary key.
        """
        return self.get_attribute(self.get_key_name())

    def get_key_name(self):
        """
        Get the primary key for the model.

        :return: The primary key name
        :rtype: str
        """
        return self.__primary_key__

    def set_key_name(self, name):
        """
        Set the primary key for the model.

        :param name: The primary key name
        :type name: str
        """
        self.__primary_key__ = name

    def get_qualified_key_name(self):
        """
        Get the table qualified key name.

        :rtype: str
        """
        return "%s.%s" % (self.get_table(), self.get_key_name())

    def uses_timestamps(self):
        """
        Determine if the model uses timestamps.

        :rtype: bool
        """
        return self.__timestamps__

    def get_morphs(self, name, type, id):
        """
        Get the polymorphic relationship columns.
        """
        if not type:
            type = name + "_type"

        if not id:
            id = name + "_id"

        return type, id

    @classmethod
    def get_morph_name(cls):
        """
        Get the name for polymorphic relations.
        """
        if not cls.__morph_name__:
            return cls._register.inverse[cls]

        return cls.__morph_name__

    def get_per_page(self):
        """
        Get the number of models to return per page.

        :rtype: int
        """
        return self._per_page

    def get_foreign_key(self):
        """
        Get the default foreign key name for the model

        :rtype: str
        """
        return "%s_id" % inflection.singularize(self.get_table())

    def get_hidden(self):
        """
        Get the hidden attributes for the model.
        """
        return self.__hidden__

    def set_hidden(self, hidden):
        """
        Set the hidden attributes for the model.

        :param hidden: The attributes to add
        :type hidden: list
        """
        self.__hidden__ = hidden

        return self

    def add_hidden(self, *attributes):
        """
        Add hidden attributes to the model.

        :param attributes: The attributes to hide
        :type attributes: list
        """
        self.__hidden__ += attributes

    def get_visible(self):
        """
        Get the visible attributes for the model.
        """
        return self.__visible__

    def set_visible(self, visible):
        """
        Set the visible attributes for the model.

        :param visible: The attributes to make visible
        :type visible: list
        """
        self.__visible__ = visible

        return self

    def add_visible(self, *attributes):
        """
        Add visible attributes to the model.

        :param attributes: The attributes to make visible
        :type attributes: list
        """
        self.__visible__ += attributes

    def get_fillable(self):
        """
        Get the fillable attributes for the model.

        :rtype: list
        """
        return self.__fillable__

    def fillable(self, fillable):
        """
        Set the fillable attributes for the model.

        :param fillable: The fillable attributes
        :type fillable: list

        :return: The current Model instance
        :rtype: Model
        """
        self.__fillable__ = fillable

        return self

    def get_guarded(self):
        """
        Get the guarded attributes.
        """
        return self.__guarded__

    def guard(self, guarded):
        """
        Set the guarded attributes.

        :param guarded: The guarded attributes
        :type guarded: list

        :return: The current Model instance
        :rtype: Model
        """
        self.__guarded__ = guarded

        return self

    @classmethod
    def unguard(cls):
        """
        Disable the mass assigment restrictions.
        """
        cls.__unguarded__ = True

    @classmethod
    def reguard(cls):
        """
        Enable the mass assignment restrictions.
        :return:
        """
        cls.__unguarded__ = False

    def is_fillable(self, key):
        """
        Determine if the given attribute can be mass assigned.

        :param key: The attribute to check
        :type key: str

        :return: Whether the attribute can be mass assigned or not
        :rtype: bool
        """
        if self.__unguarded__:
            return True

        if key in self.__fillable__:
            return True

        if self.is_guarded(key):
            return False

        return not self.__fillable__ and not key.startswith("_")

    def is_guarded(self, key):
        """
        Determine if the given attribute is guarded.

        :param key: The attribute to check
        :type key: str

        :return: Whether the attribute is guarded or not
        :rtype: bool
        """
        return key in self.__guarded__ or self.__guarded__ == ["*"]

    def totally_guarded(self):
        """
        Determine if the model is totally guarded.

        :rtype: bool
        """
        return len(self.__fillable__) == 0 and self.__guarded__ == ["*"]

    def _remove_table_from_key(self, key):
        """
        Remove the table name from a given key.

        :param key: The key to remove the table name from.
        :type key: str

        :rtype: str
        """
        if "." not in key:
            return key

        return key.split(".")[-1]

    def get_incrementing(self):
        return self.__incrementing__

    def set_incrementing(self, value):
        self.__incrementing__ = value

    def to_json(self, **options):
        """
        Convert the model instance to JSON.

        :param options: The JSON options
        :type options: dict

        :return: The JSON encoded model instance
        :rtype: str
        """
        return json.dumps(self.to_dict(), **options)

    def serialize(self):
        """
        Convert the model instance to a dictionary.

        :return: The dictionary version of the model instance
        :rtype: dict
        """
        attributes = self.attributes_to_dict()

        attributes.update(self.relations_to_dict())

        return attributes

    @deprecated
    def to_dict(self):
        """
        Convert the model instance to a dictionary.

        :return: The dictionary version of the model instance
        :rtype: dict
        """
        return self.serialize()

    def attributes_to_dict(self):
        """
        Convert the model's attributes to a dictionary.

        :rtype: dict
        """
        attributes = self._get_dictable_attributes()
        mutated_attributes = self._get_mutated_attributes()

        for key in self.get_dates():
            if not key in attributes or key in mutated_attributes:
                continue

            attributes[key] = self._format_date(attributes[key])

        for key in mutated_attributes:
            if key not in attributes:
                continue

            attributes[key] = self._mutate_attribute_for_dict(key)

        # Next we will handle any casts that have been setup for this model and cast
        # the values to their appropriate type. If the attribute has a mutator we
        # will not perform the cast on those attributes to avoid any confusion.
        for key, value in self.__casts__.items():
            if key not in attributes or key in mutated_attributes:
                continue

            attributes[key] = self._cast_attribute(key, attributes[key])

        # Here we will grab all of the appended, calculated attributes to this model
        # as these attributes are not really in the attributes array, but are run
        # when we need to array or JSON the model for convenience to the coder.
        for key in self._get_dictable_appends():
            attributes[key] = self._mutate_attribute_for_dict(key)

        return attributes

    def _get_dictable_attributes(self):
        """
        Get an attribute dictionary of all dictable attributes.

        :rtype: dict
        """
        return self._get_dictable_items(self._attributes)

    def _get_dictable_appends(self):
        """
        Get all the appendable values that are dictable.

        :rtype: list
        """
        if not self.__appends__:
            return []

        return self._get_dictable_items(dict(zip(self.__appends__, self.__appends__)))

    def relations_to_dict(self):
        """
        Get the model's relationships in dictionary form.

        :rtype: dict
        """
        attributes = {}

        for key, value in self._get_dictable_relations().items():
            if key in self.get_hidden():
                continue

            relation = None
            if hasattr(value, "serialize"):
                relation = value.serialize()
            elif hasattr(value, "to_dict"):
                relation = value.to_dict()
            elif value is None:
                relation = value

            if relation is not None or value is None:
                attributes[key] = relation

        return attributes

    def _get_dictable_relations(self):
        """
        Get an attribute dict of all dictable relations.
        """
        return self._get_dictable_items(self._relations)

    def _get_dictable_items(self, values):
        """
        Get an attribute dictionary of all dictable values.

        :param values: The values to check
        :type values: dict

        :rtype: dict
        """
        if len(self.__visible__) > 0:
            return {x: values[x] for x in values.keys() if x in self.__visible__}

        return {
            x: values[x]
            for x in values.keys()
            if x not in self.__hidden__ and not x.startswith("_")
        }

    def get_attribute(self, key, original=None):
        """
        Get an attribute from the model.

        :param key: The attribute to get
        :type key: str
        """
        in_attributes = key in self._attributes

        if in_attributes:
            return self._get_attribute_value(key)

        if key in self._relations:
            return self._relations[key]

        relation = original or super(Model, self).__getattribute__(key)

        if relation:
            return self._get_relationship_from_method(key, relation)

        raise AttributeError(key)

    def get_raw_attribute(self, key):
        """
        Get the raw underlying attribute.

        :param key: The attribute to get
        :type key: str
        """
        return self._attributes[key]

    def _get_attribute_value(self, key):
        """
        Get a plain attribute.

        :param key: The attribute to get
        :type key: str
        """
        value = self._get_attribute_from_dict(key)

        if self._has_cast(key):
            value = self._cast_attribute(key, value)
        elif key in self.get_dates():
            if value is not None:
                return self.as_datetime(value)

        return value

    def _get_attribute_from_dict(self, key):
        return self._attributes.get(key)

    def _get_relationship_from_method(self, method, relations=None):
        """
        Get a relationship value from a method.

        :param method: The method name
        :type method: str

        :rtype: mixed
        """
        relations = relations or super(Model, self).__getattribute__(method)

        if not isinstance(relations, Relation):
            raise RuntimeError(
                "Relationship method must return an object of type Relation"
            )

        self._relations[method] = relations

        return self._relations[method]

    def has_get_mutator(self, key):
        """
        Determine if a get mutator exists for an attribute.

        :param key: The attribute name
        :type key: str

        :rtype: bool
        """
        return hasattr(self, "get_%s_attribute" % inflection.underscore(key))

    def _mutate_attribute_for_dict(self, key):
        """
        Get the value of an attribute using its mutator for dict conversion.

        :param key: The attribute name
        :type key: str
        """
        value = getattr(self, key)

        if hasattr(value, "to_dict"):
            return value.to_dict()

        if key in self.get_dates():
            return self._format_date(value)

        return value

    def _has_cast(self, key):
        """
        Determine whether an attribute should be casted to a native type.

        :param key: The attribute to check
        :type key: str

        :rtype: bool
        """
        return key in self.__casts__

    def _has_set_mutator(self, key):
        """
        Determine whether an attribute has a set mutator.

        :param key: The attribute
        :type key: str

        :rtype: bool
        """
        klass = self.__class__
        if key not in self._mutator_cache[klass]:
            return False

        return self._mutator_cache[klass][key].mutator is not None

    def _is_json_castable(self, key):
        """
        Determine whether a value is JSON castable.

        :param key: The key to check
        :type key: str

        :rtype: bool
        """
        if self._has_cast(key):
            type = self._get_cast_type(key)

            return type in ["list", "dict", "json", "object"]

        return False

    def _get_cast_type(self, key):
        """
        Get the type of the cast for a model attribute.

        :param key: The attribute to get the cast for
        :type key: str

        :rtype: str
        """
        return self.__casts__[key].lower().strip()

    def _cast_attribute(self, key, value):
        """
        Cast an attribute to a native Python type

        :param key: The attribute key
        :type key: str

        :param value: The attribute value
        :type value: The attribute value

        :rtype: mixed
        """
        if value is None:
            return None

        type = self._get_cast_type(key)
        if type in ["int", "integer"]:
            return int(value)
        elif type in ["real", "float", "double"]:
            return float(value)
        elif type in ["string", "str"]:
            return str(value)
        elif type in ["bool", "boolean"]:
            return bool(value)
        elif type in ["dict", "list", "json"] and isinstance(value, basestring):
            return json.loads(value)
        else:
            return value

    def get_dates(self):
        """
        Get the attributes that should be converted to dates.

        :rtype: list
        """
        defaults = [self.CREATED_AT, self.UPDATED_AT]

        return self.__dates__ + defaults

    def from_datetime(self, value):
        """
        Convert datetime to a storable string.
        
        :param value: The datetime value
        :type value: pendulum.Pendulum or datetime.date or datetime.datetime

        :rtype: str
        """
        date_format = self.get_connection().get_query_grammar().get_date_format()

        if isinstance(value, pendulum.Pendulum):
            return value.format(date_format)

        if isinstance(value, datetime.date) and not isinstance(
            value, (datetime.datetime)
        ):
            value = pendulum.date.instance(value)

            return value.format(date_format)

        return pendulum.instance(value).format(date_format)

    def as_datetime(self, value):
        """
        Return a timestamp as a datetime.

        :rtype: pendulum.Pendulum or pendulum.Date
        """
        if isinstance(value, basestring):
            return pendulum.parse(value)

        if isinstance(value, (int, float)):
            return pendulum.from_timestamp(value)

        if isinstance(value, datetime.date) and not isinstance(
            value, (datetime.datetime)
        ):
            return pendulum.date.instance(value)

        return pendulum.instance(value)

    def get_date_format(self):
        """
        Get the format to use for timestamps and dates.

        :rtype: str
        """
        return "iso"

    def _format_date(self, date):
        """
        Format a date or timestamp.

        :param date: The date or timestamp
        :type date: datetime.datetime or datetime.date or pendulum.Pendulum

        :rtype: str
        """
        if date is None:
            return date

        format = self.get_date_format()

        if format == "iso":
            if isinstance(date, basestring):
                return pendulum.parse(date).isoformat()

            return date.isoformat()
        else:
            if isinstance(date, basestring):
                return pendulum.parse(date).format(format)

            return date.strftime(format)

    def set_attribute(self, key, value):
        """
        Set a given attribute on the model.
        """
        if self._has_set_mutator(key):
            return super(Model, self).__setattr__(key, value)

        if key in self.get_dates() and value:
            value = self.from_datetime(value)

        if self._is_json_castable(key):
            value = json.dumps(value)

        self._attributes[key] = value

    def replicate(self, except_=None):
        """
        Clone the model into a new, non-existing instance.

        :param except_: The attributes that should not be cloned
        :type except_: list

        :rtype: Model
        """
        if except_ is None:
            except_ = [
                self.get_key_name(),
                self.get_created_at_column(),
                self.get_updated_at_column(),
            ]

            attributes = {
                x: self._attributes[x] for x in self._attributes if x not in except_
            }

            instance = self.new_instance(attributes)

            instance.set_relations(dict(**self._relations))

            return instance

    def get_attributes(self):
        """
        Get all of the current attributes on the model.

        :rtype: dict
        """
        return self._attributes

    def set_raw_attributes(self, attributes, sync=False):
        """
        Set the dictionary of model attributes. No checking is done.

        :param attributes: The model attributes
        :type attributes: dict

        :param sync: Whether to sync the attributes or not
        :type sync: bool
        """
        self._attributes = dict(attributes.items())

        if sync:
            self.sync_original()

    def set_raw_attribute(self, key, value, sync=False):
        """
        Set an attribute. No checking is done.

        :param key: The attribute name
        :type key: str

        :param value: The attribute value
        :type value: mixed

        :param sync: Whether to sync the attributes or not
        :type sync: bool
        """
        self._attributes[key] = value

        if sync:
            self.sync_original()

    def get_original(self, key=None, default=None):
        """
        Get the original values

        :param key: The original key to get
        :type key: str

        :param default: The default value if the key does not exist
        :type default: mixed

        :rtype: mixed
        """
        if key is None:
            return self._original

        return self._original.get(key, default)

    def sync_original(self):
        """
        Sync the original attributes with the current.

        :rtype: Builder
        """
        self._original = dict(self._attributes.items())

        return self

    def sync_original_attribute(self, attribute):
        """
        Sync a single original attribute with its current value.

        :param attribute: The attribute to sync
        :type attribute: str

        :rtype: Model
        """
        self._original[attribute] = self._attributes[attribute]

        return self

    def is_dirty(self, *attributes):
        """
        Determine if the model or given attributes have been modified.

        :param attributes: The attributes to check
        :type attributes: list

        :rtype: boolean
        """
        dirty = self.get_dirty()

        if not attributes:
            return len(dirty) > 0

        for attribute in attributes:
            if attribute in dirty:
                return True

        return False

    def get_dirty(self):
        """
        Get the attribute that have been change since last sync.

        :rtype: list
        """
        dirty = {}

        for key, value in self._attributes.items():
            if key not in self._original:
                dirty[key] = value
            elif value != self._original[key]:
                dirty[key] = value

        return dirty

    @property
    def exists(self):
        return self._exists

    def set_exists(self, exists):
        self._exists = exists

    def set_appends(self, appends):
        """
        Sets the appendable attributes.

        :param appends: The appendable attributes
        :type appends: list
        """
        self.__appends__ = appends

        return self

    def get_relations(self):
        """
        Get all the loaded relations for the instance.

        :rtype: dict
        """
        return self._relations

    def get_relation(self, relation):
        """
        Get a specific relation.

        :param relation: The name of the relation.
        :type relation: str

        :rtype: mixed
        """
        return self._relations[relation]

    def set_relation(self, relation, value):
        """
        Set the specific relation in the model.

        :param relation: The name of the relation
        :type relation: str

        :param value: The relation
        :type value: mixed

        :return: The current Model instance
        :rtype: Model
        """
        self._relations[relation] = value

        return self

    def set_relations(self, relations):
        self._relations = relations

        return self

    def get_connection(self):
        """
        Get the database connection for the model

        :rtype: orator.connections.Connection
        """
        return self.resolve_connection(self.__connection__)

    def get_connection_name(self):
        """
        Get the database connection name for the model.

        :rtype: str
        """
        return self.__connection__

    def set_connection(self, name):
        """
        Set the connection associated with the model.

        :param name: The connection name
        :type name: str

        :return: The current model instance
        :rtype: Model
        """
        self.__connection__ = name

        return self

    @classmethod
    def resolve_connection(cls, connection=None):
        """
        Resolve a connection instance.

        :param connection: The connection name
        :type connection: str

        :rtype: orator.connections.Connection
        """
        return cls.__resolver.connection(connection)

    @classmethod
    def get_connection_resolver(cls):
        """
        Get the connection resolver instance.
        """
        return cls.__resolver

    @classmethod
    def set_connection_resolver(cls, resolver):
        """
        Set the connection resolver instance.
        """
        cls.__resolver = resolver

    @classmethod
    def unset_connection_resolver(cls):
        """
        Unset the connection resolver instance.
        """
        cls._resolver = None

    def _get_mutated_attributes(self):
        """
        Get the mutated attributes.

        :return: list
        """
        klass = self.__class__

        if klass in self._accessor_cache:
            return self._accessor_cache[klass]

        return []

    def __getattr__(self, item):
        return self.get_attribute(item)

    def __setattr__(self, key, value):
        if key in [
            "_attributes",
            "_exists",
            "_relations",
            "_original",
        ] or key.startswith("__"):
            return object.__setattr__(self, key, value)

        if self._has_set_mutator(key):
            return self.set_attribute(key, value)

        try:
            if object.__getattribute__(self, key):
                return object.__setattr__(self, key, value)
        except AttributeError:
            pass

        if callable(getattr(self, key, None)):
            return super(Model, self).__setattr__(key, value)
        else:
            self.set_attribute(key, value)

    def __delattr__(self, item):
        try:
            super(Model, self).__delattr__(item)
        except AttributeError:
            del self._attributes[item]

    def __getstate__(self):
        return {
            "attributes": self._attributes,
            "relations": self._relations,
            "exists": self._exists,
        }

    def __setstate__(self, state):
        self._boot_if_not_booted()

        self.set_raw_attributes(state["attributes"], True)
        self.set_relations(state["relations"])
        self.set_exists(state["exists"])



# -*- coding: utf-8 -*-

from ..utils import PY2, decode
from .connection import Connection
from ..query.processors.sqlite_processor import SQLiteQueryProcessor
from ..query.grammars.sqlite_grammar import SQLiteQueryGrammar
from ..schema.grammars.sqlite_grammar import SQLiteSchemaGrammar
from ..dbal.sqlite_schema_manager import SQLiteSchemaManager


class SQLiteConnection(Connection):

    name = "sqlite"

    def get_default_query_grammar(self):
        return self.with_table_prefix(SQLiteQueryGrammar())

    def get_default_post_processor(self):
        return SQLiteQueryProcessor()

    def get_default_schema_grammar(self):
        return self.with_table_prefix(SQLiteSchemaGrammar(self))

    def get_schema_manager(self):
        return SQLiteSchemaManager(self)

    def begin_transaction(self):
        self._connection.isolation_level = "DEFERRED"

        super(SQLiteConnection, self).begin_transaction()

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()
            self._connection.isolation_level = None

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
            self._connection.isolation_level = None
        else:
            self._transactions -= 1

    def prepare_bindings(self, bindings):
        bindings = super(SQLiteConnection, self).prepare_bindings(bindings)

        if PY2:
            return map(lambda x: decode(x) if isinstance(x, str) else x, bindings)

        return bindings



# -*- coding: utf-8 -*-

from .identifier import Identifier


class ColumnDiff(object):
    def __init__(
        self, old_column_name, column, changed_properties=None, from_column=None
    ):
        self.old_column_name = old_column_name
        self.column = column
        self.changed_properties = changed_properties
        self.from_column = from_column

    def has_changed(self, property_name):
        return property_name in self.changed_properties

    def get_old_column_name(self):
        return Identifier(self.old_column_name)



# -*- coding: utf-8 -*-

from ..support.fluent import Fluent


class Blueprint(object):
    def __init__(self, table):
        """
        :param table: The table to operate on
        :type table: str
        """
        self._table = table
        self._columns = []
        self._commands = []
        self.engine = None
        self.charset = None
        self.collation = None

    def build(self, connection, grammar):
        """
        Execute the blueprint against the database.

        :param connection: The connection to use
        :type connection: orator.connections.Connection

        :param grammar: The grammar to user
        :type grammar: orator.query.grammars.QueryGrammar
        """
        for statement in self.to_sql(connection, grammar):
            connection.statement(statement)

    def to_sql(self, connection, grammar):
        """
        Get the raw SQL statements for the blueprint.

        :param connection: The connection to use
        :type connection: orator.connections.Connection

        :param grammar: The grammar to user
        :type grammar: orator.schema.grammars.SchemaGrammar

        :rtype: list
        """
        self._add_implied_commands()

        statements = []

        for command in self._commands:
            method = "compile_%s" % command.name

            if hasattr(grammar, method):
                sql = getattr(grammar, method)(self, command, connection)
                if sql is not None:
                    if isinstance(sql, list):
                        statements += sql
                    else:
                        statements.append(sql)

        return statements

    def _add_implied_commands(self):
        """
        Add the commands that are implied by the blueprint.
        """
        if len(self.get_added_columns()) and not self._creating():
            self._commands.insert(0, self._create_command("add"))

        if len(self.get_changed_columns()) and not self._creating():
            self._commands.insert(0, self._create_command("change"))

        return self._add_fluent_indexes()

    def _add_fluent_indexes(self):
        """
        Add the index commands fluently specified on columns:
        """
        for column in self._columns:
            for index in ["primary", "unique", "index"]:
                column_index = column.get(index)

                if column_index is True:
                    getattr(self, index)(column.name)

                    break
                elif column_index:
                    getattr(self, index)(column.name, column_index)

                    break

    def _creating(self):
        """
        Determine if the blueprint has a create command.

        :rtype: bool
        """
        for command in self._commands:
            if command.name == "create":
                return True

        return False

    def create(self):
        """
        Indicates that the table needs to be created.

        :rtype: Fluent
        """
        return self._add_command("create")

    def drop(self):
        """
        Indicates that the table needs to be dropped.

        :rtype: Fluent
        """
        self._add_command("drop")

        return self

    def drop_if_exists(self):
        """
        Indicates that the table should be dropped if it exists.

        :rtype: Fluent
        """
        return self._add_command("drop_if_exists")

    def drop_column(self, *columns):
        """
        Indicates that the given columns should be dropped.

        :param columns: The columns to drop
        :type columns: tuple

        :rtype: Fluent
        """
        columns = list(columns)

        return self._add_command("drop_column", columns=columns)

    def rename_column(self, from_, to):
        """
        Indicates that the given columns should be renamed.

        :param from_: The original column name
        :type from_: str
        :param to: The new name of the column
        :type to: str

        :rtype: Fluent
        """
        return self._add_command("rename_column", **{"from_": from_, "to": to})

    def drop_primary(self, index=None):
        """
        Indicate that the given primary key should be dropped.

        :param index: The index
        :type index: str

        :rtype: dict
        """
        return self._drop_index_command("drop_primary", "primary", index)

    def drop_unique(self, index):
        """
        Indicate that the given unique key should be dropped.

        :param index: The index
        :type index: str

        :rtype: Fluent
        """
        return self._drop_index_command("drop_unique", "unique", index)

    def drop_index(self, index):
        """
        Indicate that the given index should be dropped.

        :param index: The index
        :type index: str

        :rtype: Fluent
        """
        return self._drop_index_command("drop_index", "index", index)

    def drop_foreign(self, index):
        """
        Indicate that the given foreign key should be dropped.

        :param index: The index
        :type index: str

        :rtype: dict
        """
        return self._drop_index_command("drop_foreign", "foreign", index)

    def drop_timestamps(self):
        """
        Indicate that the timestamp columns should be dropped.

        :rtype: Fluent
        """
        return self.drop_column("created_at", "updated_at")

    def drop_soft_deletes(self):
        """
        Indicate that the soft delete column should be dropped

        :rtype: Fluent
        """
        return self.drop_column("deleted_at")

    def rename(self, to):
        """
        Rename the table to a given name

        :param to: The new table name
        :type to: str

        :rtype: Fluent
        """
        return self._add_command("rename", to=to)

    def primary(self, columns, name=None):
        """
        Specify the primary key(s) for the table

        :param columns: The primary key(s) columns
        :type columns: str or list

        :param name: The name of the primary key
        :type name: str

        :rtype: Fluent
        """
        return self._index_command("primary", columns, name)

    def unique(self, columns, name=None):
        """
        Specify a unique index on the table

        :param columns: The primary key(s) columns
        :type columns: str or list

        :param name: The name of the primary key
        :type name: str

        :rtype: Fluent
        """
        return self._index_command("unique", columns, name)

    def index(self, columns, name=None):
        """
        Specify an index on the table

        :param columns: The primary key(s) columns
        :type columns: str or list

        :param name: The name of the primary key
        :type name: str

        :rtype: Fluent
        """
        return self._index_command("index", columns, name)

    def foreign(self, columns, name=None):
        """
        Specify an foreign key on the table

        :param columns: The foreign key(s) columns
        :type columns: str or list

        :param name: The name of the foreign key
        :type name: str

        :rtype: Fluent
        """
        return self._index_command("foreign", columns, name)

    def increments(self, column):
        """
        Create a new auto-incrementing integer column on the table.

        :param column: The auto-incrementing column
        :type column: str

        :rtype: Fluent
        """
        return self.unsigned_integer(column, True)

    def big_increments(self, column):
        """
        Create a new auto-incrementing big integer column on the table.

        :param column: The auto-incrementing column
        :type column: str

        :rtype: Fluent
        """
        return self.unsigned_big_integer(column, True)

    def char(self, column, length=255):
        """
        Create a new char column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("char", column, length=length)

    def string(self, column, length=255):
        """
        Create a new string column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("string", column, length=length)

    def text(self, column):
        """
        Create a new text column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("text", column)

    def medium_text(self, column):
        """
        Create a new medium text column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("medium_text", column)

    def long_text(self, column):
        """
        Create a new long text column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("long_text", column)

    def integer(self, column, auto_increment=False, unsigned=False):
        """
        Create a new integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :type unsigned: bool

        :rtype: Fluent
        """
        return self._add_column(
            "integer", column, auto_increment=auto_increment, unsigned=unsigned
        )

    def big_integer(self, column, auto_increment=False, unsigned=False):
        """
        Create a new big integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :type unsigned: bool

        :rtype: Fluent
        """
        return self._add_column(
            "big_integer", column, auto_increment=auto_increment, unsigned=unsigned
        )

    def medium_integer(self, column, auto_increment=False, unsigned=False):
        """
        Create a new medium integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :type unsigned: bool

        :rtype: Fluent
        """
        return self._add_column(
            "medium_integer", column, auto_increment=auto_increment, unsigned=unsigned
        )

    def tiny_integer(self, column, auto_increment=False, unsigned=False):
        """
        Create a new tiny integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :type unsigned: bool

        :rtype: Fluent
        """
        return self._add_column(
            "tiny_integer", column, auto_increment=auto_increment, unsigned=unsigned
        )

    def small_integer(self, column, auto_increment=False, unsigned=False):
        """
        Create a new small integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :type unsigned: bool

        :rtype: Fluent
        """
        return self._add_column(
            "small_integer", column, auto_increment=auto_increment, unsigned=unsigned
        )

    def unsigned_integer(self, column, auto_increment=False):
        """
        Create a new unsigned integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :rtype: Fluent
        """
        return self.integer(column, auto_increment, True)

    def unsigned_big_integer(self, column, auto_increment=False):
        """
        Create a new unsigned big integer column on the table.

        :param column: The column
        :type column: str

        :type auto_increment: bool

        :rtype: Fluent
        """
        return self.big_integer(column, auto_increment, True)

    def float(self, column, total=8, places=2):
        """
        Create a new float column on the table.

        :param column: The column
        :type column: str

        :type total: int

        :type places: 2

        :rtype: Fluent
        """
        return self._add_column("float", column, total=total, places=places)

    def double(self, column, total=None, places=None):
        """
        Create a new double column on the table.

        :param column: The column
        :type column: str

        :type total: int

        :type places: 2

        :rtype: Fluent
        """
        return self._add_column("double", column, total=total, places=places)

    def decimal(self, column, total=8, places=2):
        """
        Create a new decimal column on the table.

        :param column: The column
        :type column: str

        :type total: int

        :type places: 2

        :rtype: Fluent
        """
        return self._add_column("decimal", column, total=total, places=places)

    def boolean(self, column):
        """
        Create a new decimal column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("boolean", column)

    def enum(self, column, allowed):
        """
        Create a new enum column on the table.
        
        :param column: The column
        :type column: str
        
        :type allowed: list
        
        :rtype: Fluent
        """
        return self._add_column("enum", column, allowed=allowed)

    def json(self, column):
        """
        Create a new json column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("json", column)

    def date(self, column):
        """
        Create a new date column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("date", column)

    def datetime(self, column, with_time_zone=False):
        """
        Create a new datetime column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("datetime", column, with_time_zone=with_time_zone)

    def time(self, column, with_time_zone=False):
        """
        Create a new time column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("time", column, with_time_zone=with_time_zone)

    def timestamp(self, column, with_time_zone=False):
        """
        Create a new timestamp column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("timestamp", column, with_time_zone=with_time_zone)

    def nullable_timestamps(self):
        """
        Create nullable creation and update timestamps to the table.

        :rtype: Fluent
        """
        self.timestamp("created_at").nullable()
        self.timestamp("updated_at").nullable()

    def timestamps(self, use_current=True, with_time_zone=False):
        """
        Create creation and update timestamps to the table.

        :rtype: Fluent
        """
        if use_current:
            self.timestamp("created_at", with_time_zone=with_time_zone).use_current()
            self.timestamp("updated_at", with_time_zone=with_time_zone).use_current()
        else:
            self.timestamp("created_at", with_time_zone=with_time_zone)
            self.timestamp("updated_at", with_time_zone=with_time_zone)

    def soft_deletes(self, with_time_zone=False):
        """
        Add a "deleted at" timestamp to the table.

        :rtype: Fluent
        """
        return self.timestamp("deleted_at", with_time_zone=with_time_zone).nullable()

    def binary(self, column):
        """
        Create a new binary column on the table.

        :param column: The column
        :type column: str

        :rtype: Fluent
        """
        return self._add_column("binary", column)

    def morphs(self, name, index_name=None):
        """
        Add the proper columns for a polymorphic table.

        :type name: str

        :type index_name: str
        """
        self.unsigned_integer("%s_id" % name)
        self.string("%s_type" % name)
        self.index(["%s_id" % name, "%s_type" % name], index_name)

    def _drop_index_command(self, command, type, index):
        """
        Create a new drop index command on the blueprint.

        :param command: The command
        :type command: str

        :param type: The index type
        :type type: str

        :param index: The index name
        :type index: str

        :rtype: Fluent
        """
        columns = []

        if isinstance(index, list):
            columns = index

            index = self._create_index_name(type, columns)

        return self._index_command(command, columns, index)

    def _index_command(self, type, columns, index):
        """
        Add a new index command to the blueprint.

        :param type: The index type
        :type type: str

        :param columns: The index columns
        :type columns: list or str

        :param index: The index name
        :type index: str

        :rtype: Fluent
        """
        if not isinstance(columns, list):
            columns = [columns]

        if not index:
            index = self._create_index_name(type, columns)

        return self._add_command(type, index=index, columns=columns)

    def _create_index_name(self, type, columns):
        if not isinstance(columns, list):
            columns = [columns]

        index = "%s_%s_%s" % (
            self._table,
            "_".join([str(column) for column in columns]),
            type,
        )

        return index.lower().replace("-", "_").replace(".", "_")

    def _add_column(self, type, name, **parameters):
        """
        Add a new column to the blueprint.

        :param type: The column type
        :type type: str

        :param name: The column name
        :type name: str

        :param parameters: The column parameters
        :type parameters: dict

        :rtype: Fluent
        """
        parameters.update({"type": type, "name": name})

        column = Fluent(**parameters)
        self._columns.append(column)

        return column

    def _remove_column(self, name):
        """
        Removes a column from the blueprint.

        :param name: The column name
        :type name: str

        :rtype: Blueprint
        """
        self._columns = filter(lambda c: c.name != name, self._columns)

        return self

    def _add_command(self, name, **parameters):
        """
        Add a new command to the blueprint.

        :param name: The command name
        :type name: str

        :param parameters: The command parameters
        :type parameters: dict

        :rtype: Fluent
        """
        command = self._create_command(name, **parameters)
        self._commands.append(command)

        return command

    def _create_command(self, name, **parameters):
        """
        Create a new command.

        :param name: The command name
        :type name: str

        :param parameters: The command parameters
        :type parameters: dict

        :rtype: Fluent
        """
        parameters.update({"name": name})

        return Fluent(**parameters)

    def get_table(self):
        return self._table

    def get_columns(self):
        return self._columns

    def get_commands(self):
        return self._commands

    def get_added_columns(self):
        return list(filter(lambda column: not column.get("change"), self._columns))

    def get_changed_columns(self):
        return list(filter(lambda column: column.get("change"), self._columns))



# -*- coding: utf-8 -*-

from ..support.collection import Collection as BaseCollection


class Collection(BaseCollection):
    def load(self, *relations):
        """
        Load a set of relationships onto the collection.
        """
        if len(self.items) > 0:
            query = self.first().new_query().with_(*relations)

            self._set_items(query.eager_load_relations(self.items))

        return self

    def lists(self, value, key=None):
        """
        Get a list with the values of a given key

        :rtype: list
        """
        results = map(lambda x: getattr(x, value), self.items)

        return list(results)

    def model_keys(self):
        """
        Get the list of primary keys.

        :rtype: list
        """
        return map(lambda m: m.get_key(), self.items)



# -*- coding: utf-8 -*-

from __future__ import division
from ..utils import PY2
from .connection import Connection, run
from ..query.grammars.postgres_grammar import PostgresQueryGrammar
from ..query.processors.postgres_processor import PostgresQueryProcessor
from ..schema.grammars import PostgresSchemaGrammar
from ..dbal.postgres_schema_manager import PostgresSchemaManager


class PostgresConnection(Connection):

    name = "pgsql"

    def get_default_query_grammar(self):
        return PostgresQueryGrammar(marker=self._marker)

    def get_default_post_processor(self):
        return PostgresQueryProcessor()

    def get_default_schema_grammar(self):
        return self.with_table_prefix(PostgresSchemaGrammar(self))

    def get_schema_manager(self):
        return PostgresSchemaManager(self)

    @run
    def statement(self, query, bindings=None):
        if self.pretending():
            return True

        bindings = self.prepare_bindings(bindings)

        self._new_cursor().execute(query, bindings)

        return True

    def begin_transaction(self):
        self._reconnect_if_missing_connection()

        try:
            self._connection.autocommit = False
        except Exception as e:
            if self._caused_by_lost_connection(e):
                self.reconnect()
                self._connection.autocommit = False
            else:
                raise

        super(PostgresConnection, self).begin_transaction()

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()
            self._connection.autocommit = True

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
            self._connection.autocommit = True
        else:
            self._transactions -= 1

    def _get_cursor_query(self, query, bindings):
        if self._pretending:
            if PY2:
                return self._cursor.mogrify(query, bindings)

            return self._cursor.mogrify(query, bindings).decode()

        if not hasattr(self._cursor, "query"):
            return super(PostgresConnection, self)._get_cursor_query(query, bindings)

        if PY2:
            return self._cursor.query

        return self._cursor.query.decode()



# -*- coding: utf-8 -*-


class DBALException(Exception):

    pass


class InvalidPlatformSpecified(DBALException):
    def __init__(self, index_name, table_name):
        message = 'Invalid "platform" option specified, need to give an instance of dbal.platforms.Platform'

        super(InvalidPlatformSpecified, self).__init__(message)


class SchemaException(DBALException):

    pass


class IndexDoesNotExist(SchemaException):
    def __init__(self, index_name, table_name):
        message = 'Index "%s" does not exist on table "%s".' % (index_name, table_name)

        super(IndexDoesNotExist, self).__init__(message)


class IndexAlreadyExists(SchemaException):
    def __init__(self, index_name, table_name):
        message = 'An index with name "%s" already exists on table "%s".' % (
            index_name,
            table_name,
        )

        super(IndexAlreadyExists, self).__init__(message)


class IndexNameInvalid(SchemaException):
    def __init__(self, index_name):
        message = 'Invalid index name "%s" given, has to be [a-zA-Z0-9_]' % index_name

        super(IndexNameInvalid, self).__init__(message)


class ColumnDoesNotExist(SchemaException):
    def __init__(self, column, table_name):
        message = 'Column "%s" does not exist on table "%s".' % (column, table_name)

        super(ColumnDoesNotExist, self).__init__(message)


class ColumnAlreadyExists(SchemaException):
    def __init__(self, column, table_name):
        message = 'An column with name "%s" already exists on table "%s".' % (
            column,
            table_name,
        )

        super(ColumnAlreadyExists, self).__init__(message)


class ForeignKeyDoesNotExist(SchemaException):
    def __init__(self, constraint, table_name):
        message = 'Foreign key "%s" does not exist on table "%s".' % (
            constraint,
            table_name,
        )

        super(ForeignKeyDoesNotExist, self).__init__(message)



# -*- coding: utf-8 -*-

from collections import OrderedDict
from .table_diff import TableDiff
from .column_diff import ColumnDiff
from .index import Index
from .foreign_key_constraint import ForeignKeyConstraint


class Comparator(object):
    """
    Compares two Schemas and return an instance of SchemaDiff.
    """

    def diff_table(self, table1, table2):
        """
        Returns the difference between the tables table1 and table2.

        :type table1: Table
        :type table2: Table

        :rtype: TableDiff
        """
        changes = 0
        table_differences = TableDiff(table1.get_name())
        table_differences.from_table = table1

        table1_columns = table1.get_columns()
        table2_columns = table2.get_columns()

        # See if all the fields in table1 exist in table2
        for column_name, column in table2_columns.items():
            if not table1.has_column(column_name):
                table_differences.added_columns[column_name] = column
                changes += 1

        # See if there are any removed fields in table2
        for column_name, column in table1_columns.items():
            if not table2.has_column(column_name):
                table_differences.removed_columns[column_name] = column
                changes += 1
                continue

            # See if column has changed properties in table2
            changed_properties = self.diff_column(
                column, table2.get_column(column_name)
            )

            if changed_properties:
                column_diff = ColumnDiff(
                    column.get_name(),
                    table2.get_column(column_name),
                    changed_properties,
                )
                column_diff.from_column = column
                table_differences.changed_columns[column.get_name()] = column_diff
                changes += 1

        self.detect_column_renamings(table_differences)

        table1_indexes = table1.get_indexes()
        table2_indexes = table2.get_indexes()

        # See if all the fields in table1 exist in table2
        for index_name, index in table2_indexes.items():
            if (
                index.is_primary() and not table1.has_primary_key()
            ) or table1.has_index(index_name):
                continue

            table_differences.added_indexes[index_name] = index
            changes += 1

        # See if there are any removed fields in table2
        for index_name, index in table1_indexes.items():
            if (index.is_primary() and not table2.has_primary_key()) or (
                not index.is_primary() and not table2.has_index(index_name)
            ):
                table_differences.removed_indexes[index_name] = index
                changes += 1
                continue

            # See if index has changed in table 2
            if index.is_primary():
                table2_index = table2.get_primary_key()
            else:
                table2_index = table2.get_index(index_name)

            if self.diff_index(index, table2_index):
                table_differences.changed_indexes[index_name] = index
                changes += 1

        self.detect_index_renamings(table_differences)

        from_fkeys = OrderedDict([(k, v) for k, v in table1.get_foreign_keys().items()])
        to_fkeys = OrderedDict([(k, v) for k, v in table2.get_foreign_keys().items()])

        for key1, constraint1 in table1.get_foreign_keys().items():
            for key2, constraint2 in table2.get_foreign_keys().items():
                if self.diff_foreign_key(constraint1, constraint2) is False:
                    del from_fkeys[key1]
                    del to_fkeys[key2]
                else:
                    if constraint1.get_name().lower() == constraint2.get_name().lower():
                        table_differences.changed_foreign_keys.append(constraint2)
                        changes += 1
                        del from_fkeys[key1]
                        del to_fkeys[key2]

        for constraint1 in from_fkeys.values():
            table_differences.removed_foreign_keys.append(constraint1)
            changes += 1

        for constraint2 in to_fkeys.values():
            table_differences.added_foreign_keys.append(constraint2)
            changes += 1

        if changes:
            return table_differences

        return False

    def detect_column_renamings(self, table_differences):
        """
        Try to find columns that only changed their names.

        :type table_differences: TableDiff
        """
        rename_candidates = {}

        for added_column_name, added_column in table_differences.added_columns.items():
            for removed_column in table_differences.removed_columns.values():
                if len(self.diff_column(added_column, removed_column)) == 0:
                    if added_column.get_name() not in rename_candidates:
                        rename_candidates[added_column.get_name()] = []

                    rename_candidates[added_column.get_name()].append(
                        (removed_column, added_column, added_column_name)
                    )

        for candidate_columns in rename_candidates.values():
            if len(candidate_columns) == 1:
                removed_column, added_column, _ = candidate_columns[0]
                removed_column_name = removed_column.get_name().lower()
                added_column_name = added_column.get_name().lower()

                if removed_column_name not in table_differences.renamed_columns:
                    table_differences.renamed_columns[
                        removed_column_name
                    ] = added_column
                    del table_differences.added_columns[added_column_name]
                    del table_differences.removed_columns[removed_column_name]

    def detect_index_renamings(self, table_differences):
        """
        Try to find indexes that only changed their name,
        rename operations maybe cheaper than add/drop
        however ambiguities between different possibilities
        should not lead to renaming at all.

        :type table_differences: TableDiff
        """
        rename_candidates = OrderedDict()

        # Gather possible rename candidates by comparing
        # each added and removed index based on semantics.
        for added_index_name, added_index in table_differences.added_indexes.items():
            for removed_index in table_differences.removed_indexes.values():
                if not self.diff_index(added_index, removed_index):
                    if added_index.get_name() not in rename_candidates:
                        rename_candidates[added_index.get_name()] = []

                    rename_candidates[added_index.get_name()].append(
                        (removed_index, added_index, added_index_name)
                    )

        for candidate_indexes in rename_candidates.values():
            # If the current rename candidate contains exactly one semantically equal index,
            # we can safely rename it.
            # Otherwise it is unclear if a rename action is really intended,
            # therefore we let those ambiguous indexes be added/dropped.
            if len(candidate_indexes) == 1:
                removed_index, added_index, _ = candidate_indexes[0]

                removed_index_name = removed_index.get_name().lower()
                added_index_name = added_index.get_name().lower()

                if not removed_index_name in table_differences.renamed_indexes:
                    table_differences.renamed_indexes[removed_index_name] = added_index
                    del table_differences.added_indexes[added_index_name]
                    del table_differences.removed_indexes[removed_index_name]

    def diff_foreign_key(self, key1, key2):
        """
        :type key1: ForeignKeyConstraint

        :type key2: ForeignKeyConstraint

        :rtype: bool
        """
        key1_unquoted_local_columns = [
            c.lower() for c in key1.get_unquoted_local_columns()
        ]
        key2_unquoted_local_columns = [
            c.lower() for c in key2.get_unquoted_local_columns()
        ]

        if key1_unquoted_local_columns != key2_unquoted_local_columns:
            return True

        key1_unquoted_foreign_columns = [
            c.lower() for c in key1.get_unquoted_foreign_columns()
        ]
        key2_unquoted_foreign_columns = [
            c.lower() for c in key2.get_unquoted_foreign_columns()
        ]

        if key1_unquoted_foreign_columns != key2_unquoted_foreign_columns:
            return True

        if (
            key1.get_unqualified_foreign_table_name()
            != key2.get_unqualified_foreign_table_name()
        ):
            return True

        if key1.on_update() != key2.on_update():
            return True

        if key1.on_delete() != key2.on_delete():
            return True

        return False

    def diff_column(self, column1, column2):
        """
        Returns the difference between column1 and column2

        :type column1: orator.dbal.column.Column
        :type column2: orator.dbal.column.Column

        :rtype: list
        """
        properties1 = column1.to_dict()
        properties2 = column2.to_dict()

        changed_properties = []

        for prop in ["type", "notnull", "unsigned", "autoincrement"]:
            if properties1[prop] != properties2[prop]:
                changed_properties.append(prop)

        if (
            properties1["default"] != properties2["default"]
            or (properties1["default"] is None and properties2["default"] is not None)
            or (properties2["default"] is None and properties1["default"] is not None)
        ):
            changed_properties.append("default")

        if (
            properties1["type"] == "string"
            and properties1["type"] != "guid"
            or properties1["type"] in ["binary", "blob"]
        ):
            length1 = properties1["length"] or 255
            length2 = properties2["length"] or 255

            if length1 != length2:
                changed_properties.append("length")

            if properties1["fixed"] != properties2["fixed"]:
                changed_properties.append("fixed")
        elif properties1["type"] in ["decimal", "float", "double precision"]:
            precision1 = properties1["precision"] or 10
            precision2 = properties2["precision"] or 10

            if precision1 != precision2:
                changed_properties.append("precision")

            if properties1["scale"] != properties2["scale"]:
                changed_properties.append("scale")

        return list(set(changed_properties))

    def diff_index(self, index1, index2):
        """
        Finds the difference between the indexes index1 and index2.

        Compares index1 with index2 and returns True if there are any
        differences or False in case there are no differences.

        :type index1: Index
        :type index2: Index

        :rtype: bool
        """
        if index1.is_fullfilled_by(index2) and index2.is_fullfilled_by(index1):
            return False

        return True



# -*- coding: utf-8 -*-

from .has_one_or_many import HasOneOrMany


class MorphOneOrMany(HasOneOrMany):
    def __init__(self, query, parent, morph_type, foreign_key, local_key):
        """
        :type query: orator.orm.Builder

        :param parent: The parent model
        :type parent: Model

        :param morph_type: The type of the morph
        :type morph_type: str

        :param foreign_key: The foreign key of the parent model
        :type foreign_key: str

        :param local_key: The local key of the parent model
        :type local_key: str
        """
        self._morph_type = morph_type
        self._morph_name = parent.get_morph_name()

        super(MorphOneOrMany, self).__init__(query, parent, foreign_key, local_key)

    def add_constraints(self):
        """
        Set the base constraints of the relation query
        """
        if self._constraints:
            super(MorphOneOrMany, self).add_constraints()

            self._query.where(self._morph_type, self._morph_name)

    def get_relation_count_query(self, query, parent):
        """
        Add the constraints for a relationship count query.

        :type query: Builder
        :type parent: Builder

        :rtype: Builder
        """
        query = super(MorphOneOrMany, self).get_relation_count_query(query, parent)

        return query.where(self._morph_type, self._morph_name)

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        super(MorphOneOrMany, self).add_eager_constraints(models)

        self._query.where(self._morph_type, self._morph_name)

    def save(self, model):
        """
        Attach a model instance to the parent models.

        :param model: The model instance to attach
        :type model: Model

        :rtype: Model
        """
        model.set_attribute(self.get_plain_morph_type(), self._morph_name)

        return super(MorphOneOrMany, self).save(model)

    def find_or_new(self, id, columns=None):
        """
        Find a model by its primary key or return new instance of the related model.

        :param id: The primary key
        :type id: mixed

        :param columns:  The columns to retrieve
        :type columns: list

        :rtype: Collection or Model
        """
        if columns is None:
            columns = ["*"]

        instance = self._query.find(id, columns)

        if instance is None:
            instance = self._related.new_instance()
            self._set_foreign_attributes_for_create(instance)

        return instance

    def first_or_new(self, _attributes=None, **attributes):
        """
        Get the first related model record matching the attributes or instantiate it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._query.where(attributes).first()

        if instance is None:
            instance = self._related.new_instance()
            self._set_foreign_attributes_for_create(instance)

        return instance

    def first_or_create(self, _attributes=None, **attributes):
        """
        Get the first related record matching the attributes or create it.

        :param attributes:  The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._query.where(attributes).first()

        if instance is None:
            instance = self.create(**attributes)

        return instance

    def update_or_create(self, attributes, values=None):
        """
        Create or update a related record matching the attributes, and fill it with values.

        :param attributes: The attributes
        :type attributes: dict

        :param values: The values
        :type values: dict

        :rtype: Model
        """
        instance = self.first_or_new(**attributes)

        instance.fill(values)

        instance.save()

        return instance

    def create(self, _attributes=None, **attributes):
        """
        Create a new instance of the related model.

        :param attributes: The attributes
        :type attributes: dict

        :rtype: Model
        """
        if _attributes is not None:
            attributes.update(_attributes)

        instance = self._related.new_instance(attributes)

        self._set_foreign_attributes_for_create(instance)

        instance.save()

        return instance

    def _set_foreign_attributes_for_create(self, model):
        """
        Set the foreign ID and type for creation a related model.
        """
        model.set_attribute(self.get_plain_foreign_key(), self.get_parent_key())

        model.set_attribute(self.get_plain_morph_type(), self._morph_name)

    def get_morph_type(self):
        return self._morph_type

    def get_plain_morph_type(self):
        return self._morph_type.split(".")[-1]

    def get_morph_name(self):
        return self._morph_name

    def _new_instance(self, parent):
        return self.__class__(
            self.new_query(),
            parent,
            self._morph_type,
            self._foreign_key,
            self._local_key,
        )



# -*- coding: utf-8 -*-

from ...support.grammar import Grammar
from ...support.fluent import Fluent
from ...query.expression import QueryExpression
from ...dbal.column import Column
from ...dbal.table_diff import TableDiff
from ...dbal.comparator import Comparator
from ..blueprint import Blueprint


class SchemaGrammar(Grammar):
    def __init__(self, connection):
        super(SchemaGrammar, self).__init__(marker=connection.get_marker())

        self._connection = connection

    def compile_rename_column(self, blueprint, command, connection):
        """
        Compile a rename column command.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: orator.connections.Connection

        :rtype: list
        """
        schema = connection.get_schema_manager()

        table = self.get_table_prefix() + blueprint.get_table()

        column = connection.get_column(table, command.from_)

        table_diff = self._get_renamed_diff(blueprint, command, column, schema)

        return schema.get_database_platform().get_alter_table_sql(table_diff)

    def _get_renamed_diff(self, blueprint, command, column, schema):
        """
        Get a new column instance with the new column name.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param column: The column
        :type column: orator.dbal.Column

        :param schema: The schema
        :type schema: orator.dbal.SchemaManager

        :rtype: orator.dbal.TableDiff
        """
        table_diff = self._get_table_diff(blueprint, schema)

        return self._set_renamed_columns(table_diff, command, column)

    def _set_renamed_columns(self, table_diff, command, column):
        """
        Set the renamed columns on the table diff.

        :rtype: orator.dbal.TableDiff
        """
        new_column = Column(command.to, column.get_type(), column.to_dict())

        table_diff.renamed_columns = {command.from_: new_column}

        return table_diff

    def compile_foreign(self, blueprint, command, _):
        """
        Compile a foreign key command.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :rtype: str
        """
        table = self.wrap_table(blueprint)

        on = self.wrap_table(command.on)

        columns = self.columnize(command.columns)

        on_columns = self.columnize(
            command.references
            if isinstance(command.references, list)
            else [command.references]
        )

        sql = "ALTER TABLE %s ADD CONSTRAINT %s " % (table, command.index)

        sql += "FOREIGN KEY (%s) REFERENCES %s (%s)" % (columns, on, on_columns)

        if command.get("on_delete"):
            sql += " ON DELETE %s" % command.on_delete

        if command.get("on_update"):
            sql += " ON UPDATE %s" % command.on_update

        return sql

    def _get_columns(self, blueprint):
        """
        Get the blueprint's columns definitions.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :rtype: list
        """
        columns = []

        for column in blueprint.get_added_columns():
            sql = self.wrap(column) + " " + self._get_type(column)

            columns.append(self._add_modifiers(sql, blueprint, column))

        return columns

    def _add_modifiers(self, sql, blueprint, column):
        """
        Add the column modifiers to the deifinition
        """
        for modifier in self._modifiers:
            method = "_modify_%s" % modifier

            if hasattr(self, method):
                sql += getattr(self, method)(blueprint, column)

        return sql

    def _get_command_by_name(self, blueprint, name):
        """
        Get the primary key command it it exists.
        """
        commands = self._get_commands_by_name(blueprint, name)

        if len(commands):
            return commands[0]

    def _get_commands_by_name(self, blueprint, name):
        """
        Get all of the commands with a given name.
        """
        return list(filter(lambda value: value.name == name, blueprint.get_commands()))

    def _get_type(self, column):
        """
        Get the SQL for the column data type.

        :param column: The column
        :type column: Fluent

        :rtype sql
        """
        return getattr(self, "_type_%s" % column.type)(column)

    def prefix_list(self, prefix, values):
        """
        Add a prefix to a list of values.
        """
        return list(map(lambda value: prefix + " " + value, values))

    def wrap_table(self, table):
        if isinstance(table, Blueprint):
            table = table.get_table()

        return super(SchemaGrammar, self).wrap_table(table)

    def wrap(self, value, prefix_alias=False):
        if isinstance(value, Fluent):
            value = value.name

        return super(SchemaGrammar, self).wrap(value, prefix_alias)

    def _get_default_value(self, value):
        """
        Format a value so that it can be used in "default" clauses.
        """
        if isinstance(value, QueryExpression):
            return value

        if isinstance(value, bool):
            return "'%s'" % int(value)

        return "'%s'" % value

    def _get_table_diff(self, blueprint, schema):
        table = self.get_table_prefix() + blueprint.get_table()

        table_diff = TableDiff(table)

        table_diff.from_table = schema.list_table_details(table)

        return table_diff

    def compile_change(self, blueprint, command, connection):
        """
        Compile a change column command into a series of SQL statement.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: orator.connections.Connection

        :rtype: list
        """
        schema = connection.get_schema_manager()

        table_diff = self._get_changed_diff(blueprint, schema)

        if table_diff:
            sql = schema.get_database_platform().get_alter_table_sql(table_diff)

            if isinstance(sql, list):
                return sql

            return [sql]

        return []

    def _get_changed_diff(self, blueprint, schema):
        """
        Get the table diffrence for the given changes.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param schema: The schema
        :type schema: orator.dbal.SchemaManager

        :rtype: orator.dbal.TableDiff
        """
        table = schema.list_table_details(
            self.get_table_prefix() + blueprint.get_table()
        )

        return Comparator().diff_table(
            table, self._get_table_with_column_changes(blueprint, table)
        )

    def _get_table_with_column_changes(self, blueprint, table):
        """
        Get a copy of the given table after making the column changes.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :type table: orator.dbal.table.Table

        :rtype: orator.dbal.table.Table
        """
        table = table.clone()

        for fluent in blueprint.get_changed_columns():
            column = self._get_column_for_change(table, fluent)

            for key, value in fluent.get_attributes().items():
                option = self._map_fluent_option(key)

                if option is not None:
                    method = "set_%s" % option

                    if hasattr(column, method):
                        getattr(column, method)(self._map_fluent_value(option, value))

        return table

    def _get_column_for_change(self, table, fluent):
        """
        Get the column instance for a column change.

        :type table: orator.dbal.table.Table

        :rtype: orator.dbal.column.Column
        """
        return table.change_column(
            fluent.name, self._get_column_change_options(fluent)
        ).get_column(fluent.name)

    def _get_column_change_options(self, fluent):
        """
        Get the column change options.
        """
        options = {
            "name": fluent.name,
            "type": self._get_dbal_column_type(fluent.type),
            "default": fluent.get("default"),
        }

        if fluent.type in ["string"]:
            options["length"] = fluent.length

        return options

    def _get_dbal_column_type(self, type_):
        """
        Get the dbal column type.

        :param type_: The fluent type
        :type type_: str

        :rtype: str
        """
        type_ = type_.lower()

        if type_ == "big_integer":
            type_ = "bigint"
        elif type == "small_integer":
            type_ = "smallint"
        elif type_ in ["medium_text", "long_text"]:
            type_ = "text"

        return type_

    def _map_fluent_option(self, attribute):
        if attribute in ["type", "name"]:
            return
        elif attribute == "nullable":
            return "notnull"
        elif attribute == "total":
            return "precision"
        elif attribute == "places":
            return "scale"
        else:
            return

    def _map_fluent_value(self, option, value):
        if option == "notnull":
            return not value

        return value

    def platform_version(self, parts=2):
        return self._connection.server_version[:parts]

    def platform(self):
        """
        Returns the dbal database platform.

        :rtype: orator.dbal.platforms.platform.Platform
        """
        return self._connection.get_database_platform()



# -*- coding: utf-8 -*-

from orator.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class MigrateCommand(BaseCommand):
    """
    Run the database migrations.

    migrate
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
        {--s|seed : Indicates if the seed task should be re-run.}
        {--seed-path= : The path of seeds files to be executed.
                        Defaults to <comment>./seeders</comment>.}
        {--P|pretend : Dump the SQL queries that would be run.}
        {--f|force : Force the operation to run.}
    """

    def handle(self):
        if not self.confirm_to_proceed(
            "<question>Are you sure you want to proceed with the migration?</question> "
        ):
            return

        database = self.option("database")
        repository = DatabaseMigrationRepository(self.resolver, "migrations")

        migrator = Migrator(repository, self.resolver)

        self._prepare_database(migrator, database)

        pretend = self.option("pretend")

        path = self.option("path")

        if path is None:
            path = self._get_migration_path()

        migrator.run(path, pretend)

        for note in migrator.get_notes():
            self.line(note)

        # If the "seed" option has been given, we will rerun the database seed task
        # to repopulate the database.
        if self.option("seed"):
            options = [("--force", self.option("force"))]

            if database:
                options.append(("--database", database))

            if self.get_definition().has_option("config"):
                options.append(("--config", self.option("config")))

            if self.option("seed-path"):
                options.append(("--path", self.option("seed-path")))

            self.call("db:seed", options)

    def _prepare_database(self, migrator, database):
        migrator.set_connection(database)

        if not migrator.repository_exists():
            options = []

            if database:
                options.append(("--database", database))

            if self.get_definition().has_option("config"):
                options.append(("--config", self.option("config")))

            self.call("migrate:install", options)
